# ----------------------------------------------------------------------------
# Copyright (c) 2019--, gemelli development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# ----------------------------------------------------------------------------

import biom
import skbio
import numpy as np
import pandas as pd
from typing import Union, Optional
from skbio import TreeNode, OrdinationResults, DistanceMatrix
from gemelli.matrix_completion import MatrixCompletion
from gemelli.preprocessing import (matrix_rclr,
                                   fast_unifrac,
                                   bp_read_phylogeny,
                                   retrieve_t2t_taxonomy,
                                   create_taxonomy_metadata)
from gemelli._defaults import (DEFAULT_COMP, DEFAULT_MTD,
                               DEFAULT_MSC, DEFAULT_MFC,
                               DEFAULT_OPTSPACE_ITERATIONS,
                               DEFAULT_MFF)
from scipy.linalg import svd
# import QIIME2 if in a Q2env otherwise set type to str
try:
    from q2_types.tree import NewickFormat
except ImportError:
    # python does not check but technically this is the type
    NewickFormat = str


def phylogenetic_rpca_without_taxonomy(
        table: biom.Table,
        phylogeny: NewickFormat,
        n_components: Union[int, str] = DEFAULT_COMP,
        min_sample_count: int = DEFAULT_MSC,
        min_feature_count: int = DEFAULT_MFC,
        min_feature_frequency: float = DEFAULT_MFF,
        min_depth: int = DEFAULT_MTD,
        max_iterations: int = DEFAULT_OPTSPACE_ITERATIONS) -> (
        OrdinationResults, DistanceMatrix,
        TreeNode, biom.Table):
    """
    Runs phylogenetic RPCA without taxonomy. This code will
    be run QIIME2 versions of gemelli. Outside of QIIME2
    please use phylogenetic_rpca.
    """

    output = phylogenetic_rpca(table=table,
                               phylogeny=phylogeny,
                               n_components=n_components,
                               min_sample_count=min_sample_count,
                               min_feature_count=min_feature_count,
                               min_feature_frequency=min_feature_frequency,
                               min_depth=min_depth,
                               max_iterations=max_iterations)
    ord_res, dist_res, phylogeny, counts_by_node, _ = output

    return ord_res, dist_res, phylogeny, counts_by_node


def phylogenetic_rpca_with_taxonomy(
            table: biom.Table,
            phylogeny: NewickFormat,
            taxonomy: pd.DataFrame,
            n_components: Union[int, str] = DEFAULT_COMP,
            min_sample_count: int = DEFAULT_MSC,
            min_feature_count: int = DEFAULT_MFC,
            min_feature_frequency: float = DEFAULT_MFF,
            min_depth: int = DEFAULT_MTD,
            max_iterations: int = DEFAULT_OPTSPACE_ITERATIONS) -> (
        OrdinationResults, DistanceMatrix,
        TreeNode, biom.Table, pd.DataFrame):
    """
    Runs phylogenetic RPCA with taxonomy. This code will
    be run QIIME2 versions of gemelli. Outside of QIIME2
    please use phylogenetic_rpca.
    """
    output = phylogenetic_rpca(table=table,
                               phylogeny=phylogeny,
                               taxonomy=taxonomy,
                               n_components=n_components,
                               min_sample_count=min_sample_count,
                               min_feature_count=min_feature_count,
                               min_feature_frequency=min_feature_frequency,
                               min_depth=min_depth,
                               max_iterations=max_iterations)
    ord_res, dist_res, phylogeny, counts_by_node, result_taxonomy = output

    return ord_res, dist_res, phylogeny, counts_by_node, result_taxonomy


def phylogenetic_rpca(table: biom.Table,
                      phylogeny: NewickFormat,
                      taxonomy: Optional[pd.DataFrame] = None,
                      n_components: Union[int, str] = DEFAULT_COMP,
                      min_sample_count: int = DEFAULT_MSC,
                      min_feature_count: int = DEFAULT_MFC,
                      min_feature_frequency: float = DEFAULT_MFF,
                      min_depth: int = DEFAULT_MTD,
                      max_iterations: int = DEFAULT_OPTSPACE_ITERATIONS) -> (
                          OrdinationResults, DistanceMatrix,
                          TreeNode, biom.Table, Optional[pd.DataFrame]):
    """
    Performs robust phylogenetic center log-ratio transform and
    robust PCA. The robust PCA and enter log-ratio transform
    operate on only observed values of the data.
    For more information see (1 and 2).

    Parameters
    ----------
    table: numpy.ndarray, required
    The feature table in biom format containing the
    samples over which metric should be computed.

    phylogeny: str, required
    Path to the file containing the phylogenetic tree containing tip
    identifiers that correspond to the feature identifiers in the table.
    This tree can contain tip ids that are not present in the table,
    but all feature ids in the table must be present in this tree.

    taxonomy: pd.DataFrame, optional
    Taxonomy file in QIIME2 formatting. See the feature metdata
    section of https://docs.qiime2.org/2021.11/tutorials/metadata

    n_components: int, optional : Default is 3
    The underlying rank of the data and number of
    output dimentions.

    min_sample_count: int, optional : Default is 0
    Minimum sum cutoff of sample across all features.
    The value can be at minimum zero and must be an
    whole integer. It is suggested to be greater than
    or equal to 500.

    min_feature_count: int, optional : Default is 0
    Minimum sum cutoff of features across all samples.
    The value can be at minimum zero and must be
    an whole integer.

    min_feature_frequency: float, optional : Default is 0
    Minimum percentage of samples a feature must appear
    with a value greater than zero. This value can range
    from 0 to 100 with decimal values allowed.

    max_iterations: int, optional : Default is 5
    The number of convex iterations to optimize the solution
    If iteration is not specified, then the default iteration is 5.
    Which redcues to a satisfactory error threshold.

    Returns
    -------
    OrdinationResults
        A biplot of the (Robust Aitchison) RPCA feature loadings

    DistanceMatrix
        The Aitchison distance of the sample loadings from RPCA.

    TreeNode
        The input tree with all nodes matched in name to the
        features in the counts-by-node table

    biom.Table
        A table with all tree internal nodes as features with the
        sum of all children of that node (i.e. FastUniFrac).

    Optional[pd.DataFrame]
        The resulting tax2Tree taxonomy and will include taxonomy for both
        internal nodes and tips. Note: this will only be output
        if taxonomy was given as input.

    Raises
    ------
    ValueError
        `ValueError: n_components must be at least 2`.

    ValueError
        `ValueError: max_iterations must be at least 1`.

    ValueError
        `ValueError: Data-table contains either np.inf or -np.inf`.

    ValueError
        `ValueError: The n_components must be less
            than the minimum shape of the input table`.

    References
    ----------
    .. [1] Martino C, Morton JT, Marotz CA, Thompson LR, Tripathi A,
           Knight R, Zengler K. 2019. A Novel Sparse Compositional
           Technique Reveals Microbial Perturbations. mSystems 4.
    .. [2] Keshavan RH, Oh S, Montanari A. 2009. Matrix completion
            from a few entries (2009_ IEEE International
            Symposium on Information Theory

    Examples
    --------
    import numpy as np
    import pandas as pd
    from biom import Table
    from gemelli.rpca import phylogenetic_rpca

    # make a table
    X = np.array([[9, 3, 0, 0],
                [9, 9, 0, 1],
                [0, 1, 4, 5],
                [0, 0, 3, 4],
                [1, 0, 8, 9]])
    sample_ids = ['s1','s2','s3','s4']
    feature_ids = ['f1','f2','f3','f4','f5']
    bt = Table(X, feature_ids, sample_ids)
    # write an example tree to read
    f = open("demo-tree.nwk", "w")
    newick = '(((f1:1,f2:1)n9:1,f3:1)n8:1,(f4:1,f5:1)n2:1)n1:1;'
    f.write(newick)
    f.close()
    # run RPCA without taxonomy
    # s1/s2 will seperate from s3/s4
    (ordination, distance_matrix,
    tree, phylo_table, _) = phylogenetic_rpca(bt, 'demo-tree.nwk')

    # make mock taxonomy
    taxonomy = pd.DataFrame({fid:['k__kingdom; p__phylum;'
                                'c__class; o__order; '
                                'f__family; g__genus;'
                                's__',
                                0.99]
                            for fid in feature_ids},
                            ['Taxon', 'Confidence']).T
    # run RPCA with taxonomy
    # s1/s2 will seperate from s3/s4
    (ordination, distance_matrix,
    tree, phylo_table,
    lca_taxonomy) = phylogenetic_rpca(bt, 'demo-tree.nwk', taxonomy)

    """

    # use helper to process table
    table = rpca_table_processing(table,
                                  min_sample_count,
                                  min_feature_count,
                                  min_feature_frequency)

    # import the tree based on filtered table
    phylogeny = bp_read_phylogeny(table,
                                  phylogeny,
                                  min_depth)
    # build the vectorized table
    counts_by_node, tree_index, branch_lengths, fids, otu_ids\
        = fast_unifrac(table, phylogeny)
    # Robust-clt (matrix_rclr) preprocessing
    rclr_table = matrix_rclr(counts_by_node, branch_lengths=branch_lengths)
    # run OptSpace (RPCA)
    ord_res, dist_res = optspace_helper(rclr_table, fids, table.ids(),
                                        n_components=n_components)
    # import expanded table
    counts_by_node = biom.Table(counts_by_node.T,
                                fids, table.ids())
    result_taxonomy = None
    if taxonomy is not None:
        # collect taxonomic information for all tree nodes.
        traversed_taxonomy = retrieve_t2t_taxonomy(phylogeny, taxonomy)
        result_taxonomy = create_taxonomy_metadata(phylogeny,
                                                   traversed_taxonomy)

    return ord_res, dist_res, phylogeny, counts_by_node, result_taxonomy


def rpca(table: biom.Table,
         n_components: Union[int, str] = DEFAULT_COMP,
         min_sample_count: int = DEFAULT_MSC,
         min_feature_count: int = DEFAULT_MFC,
         min_feature_frequency: float = DEFAULT_MFF,
         max_iterations: int = DEFAULT_OPTSPACE_ITERATIONS) -> (
        OrdinationResults,
        DistanceMatrix):
    """
    Performs robust center log-ratio transform and
    robust PCA. The robust PCA and enter log-ratio transform
    operate on only observed values of the data.
    For more information see (1 and 2).

    Parameters
    ----------
    table: numpy.ndarray, required
    The feature table in biom format containing the
    samples over which metric should be computed.

    n_components: int, optional : Default is 3
    The underlying rank of the data and number of
    output dimentions.

    min_sample_count: int, optional : Default is 0
    Minimum sum cutoff of sample across all features.
    The value can be at minimum zero and must be an
    whole integer. It is suggested to be greater than
    or equal to 500.

    min_feature_count: int, optional : Default is 0
    Minimum sum cutoff of features across all samples.
    The value can be at minimum zero and must be
    an whole integer.

    min_feature_frequency: float, optional : Default is 0
    Minimum percentage of samples a feature must appear
    with a value greater than zero. This value can range
    from 0 to 100 with decimal values allowed.

    max_iterations: int, optional : Default is 5
    The number of convex iterations to optimize the solution
    If iteration is not specified, then the default iteration is 5.
    Which redcues to a satisfactory error threshold.

    Returns
    -------
    OrdinationResults
        A biplot of the (Robust Aitchison) RPCA feature loadings

    DistanceMatrix
        The Aitchison distance of the sample loadings from RPCA.

    Raises
    ------
    ValueError
        `ValueError: n_components must be at least 2`.

    ValueError
        `ValueError: max_iterations must be at least 1`.

    ValueError
        `ValueError: Data-table contains either np.inf or -np.inf`.

    ValueError
        `ValueError: The n_components must be less
            than the minimum shape of the input table`.

    References
    ----------
    .. [1] Martino C, Morton JT, Marotz CA, Thompson LR, Tripathi A,
           Knight R, Zengler K. 2019. A Novel Sparse Compositional
           Technique Reveals Microbial Perturbations. mSystems 4.
    .. [2] Keshavan RH, Oh S, Montanari A. 2009. Matrix completion
            from a few entries (2009_ IEEE International
            Symposium on Information Theory

    Examples
    --------
    import numpy as np
    from biom import Table
    from gemelli.rpca import rpca

    # make a table
    X = np.array([[9, 3, 0, 0],
                [9, 9, 0, 1],
                [0, 1, 4, 5],
                [0, 0, 3, 4],
                [1, 0, 8, 9]])
    sample_ids = ['s1','s2','s3','s4']
    feature_ids = ['f1','f2','f3','f4','f5']
    bt = Table(X, feature_ids, sample_ids)
    # run RPCA (s1/s2 will seperate from s3/s4)
    ordination, distance_matrix = rpca(bt)

    """
    # use helper to process table
    table = rpca_table_processing(table,
                                  min_sample_count,
                                  min_feature_count,
                                  min_feature_frequency)
    # Robust-clt (matrix_rclr) preprocessing
    rclr_table = matrix_rclr(table.matrix_data.toarray().T)
    # run OptSpace (RPCA)
    ord_res, dist_res = optspace_helper(rclr_table,
                                        table.ids('observation'),
                                        table.ids(), n_components=n_components)

    return ord_res, dist_res


def optspace_helper(rclr_table: np.array,
                    feature_ids: list,
                    subject_ids: list,
                    n_components: Union[int, str] = DEFAULT_COMP,
                    max_iterations: int = DEFAULT_OPTSPACE_ITERATIONS) -> (
                        OrdinationResults,
                        DistanceMatrix):
    """
    Helper function. Please use rpca directly.
    """
    # run OptSpace (RPCA)
    opt = MatrixCompletion(n_components=n_components,
                           max_iterations=max_iterations).fit(rclr_table)
    # get new n-comp when applicable
    n_components = opt.s.shape[0]
    # get PC column labels for the skbio OrdinationResults
    rename_cols = ['PC' + str(i + 1) for i in range(n_components)]
    # get completed matrix for centering
    X = opt.sample_weights @ opt.s @ opt.feature_weights.T
    # center again around zero after completion
    X = X - X.mean(axis=0)
    X = X - X.mean(axis=1).reshape(-1, 1)
    # re-factor the data
    u, s, v = svd(X)
    # only take n-components
    u = u[:, :n_components]
    v = v.T[:, :n_components]
    # calc. the new variance using projection
    p = s**2 / np.sum(s**2)
    p = p[:n_components]
    s = s[:n_components]
    # save the loadings
    feature_loading = pd.DataFrame(v, index=feature_ids,
                                   columns=rename_cols)
    sample_loading = pd.DataFrame(u, index=subject_ids,
                                  columns=rename_cols)
    # % var explained
    proportion_explained = pd.Series(p, index=rename_cols)
    # get eigenvalues
    eigvals = pd.Series(s, index=rename_cols)

    # if the n_components is two add PC3 of zeros
    # this is referenced as in issue in
    # <https://github.com/biocore/emperor/commit
    # /a93f029548c421cb0ba365b4294f7a5a6b0209ce>
    # discussed in gemelli -- PR#29
    if n_components == 2:
        feature_loading['PC3'] = [0] * len(feature_loading.index)
        sample_loading['PC3'] = [0] * len(sample_loading.index)
        eigvals.loc['PC3'] = 0
        proportion_explained.loc['PC3'] = 0

    # save ordination results
    short_method_name = 'rpca_biplot'
    long_method_name = '(Robust Aitchison) RPCA Biplot'
    ord_res = skbio.OrdinationResults(
        short_method_name,
        long_method_name,
        eigvals.copy(),
        samples=sample_loading.copy(),
        features=feature_loading.copy(),
        proportion_explained=proportion_explained.copy())
    # save distance matrix
    dist_res = DistanceMatrix(opt.distance, ids=sample_loading.index)

    return ord_res, dist_res


def auto_rpca(table: biom.Table,
              min_sample_count: int = DEFAULT_MSC,
              min_feature_count: int = DEFAULT_MFC,
              min_feature_frequency: float = DEFAULT_MFF,
              max_iterations: int = DEFAULT_OPTSPACE_ITERATIONS) -> (
        OrdinationResults,
        DistanceMatrix):
    """Runs RPCA but with auto estimation of the
       rank peramater.
    """
    ord_res, dist_res = rpca(table,
                             n_components='auto',
                             min_sample_count=min_sample_count,
                             min_feature_count=min_feature_count,
                             min_feature_frequency=min_feature_frequency,
                             max_iterations=max_iterations)
    return ord_res, dist_res


def rpca_table_processing(table: biom.Table,
                          min_sample_count: int = DEFAULT_MSC,
                          min_feature_count: int = DEFAULT_MFC,
                          min_feature_frequency: float = DEFAULT_MFF) -> (
                              biom.Table):
    """Filter and checks the table validity for RPCA.
    """
    # get shape of table
    n_features, n_samples = table.shape

    # filter sample to min seq. depth
    def sample_filter(val, id_, md):
        return sum(val) > min_sample_count

    # filter features to min total counts
    def observation_filter(val, id_, md):
        return sum(val) > min_feature_count

    # filter features by N samples presence
    def frequency_filter(val, id_, md):
        return (np.sum(val > 0) / n_samples) > (min_feature_frequency / 100)

    # filter and import table for each filter above
    table = table.filter(observation_filter, axis='observation')
    table = table.filter(frequency_filter, axis='observation')
    table = table.filter(sample_filter, axis='sample')

    # check the table after filtering
    if len(table.ids()) != len(set(table.ids())):
        raise ValueError('Data-table contains duplicate sample IDs')
    if len(table.ids('observation')) != len(set(table.ids('observation'))):
        raise ValueError('Data-table contains duplicate feature IDs')

    return table



from gemelli.optspace import (
    grassmann_manifold_one,
    cost_function,
    gradient_decent,
    grassmann_manifold_two,
    line_search,
    singular_values,
    OptSpace,
    svd_sort,
    rank_estimate)
import numpy as np
from numpy.linalg import norm
import unittest
import numpy.testing as npt
from scipy.io import loadmat
from skbio.util import get_data_path
from numpy.testing import assert_array_almost_equal


class TestOptspace(unittest.TestCase):
    def setUp(self):
        pass

    def test_rank_estimation(self):
        """Test rank estimation is accurate."""
        N = 100
        D = 5000
        k = 3
        U = np.random.standard_normal(size=(N, k))
        V = np.random.standard_normal(size=(k, D))
        Y = U @ V
        # randomly mask Y
        mask = np.random.random(size=(N, D))
        Y[mask > .5] = 0
        # get eps
        total_nonzeros = np.count_nonzero(Y)
        eps = total_nonzeros / np.sqrt(N * D)
        # estimate
        self.assertEqual(k, rank_estimate(Y, eps))

    def test_G(self):
        """Test first grassmann manifold runs."""
        X = np.ones((10, 10))
        m0 = 2
        r = 2
        exp = grassmann_manifold_one(X, m0, r)
        self.assertAlmostEqual(exp, 0.644944589179)

    def test_G_z_0(self):
        """Test first grassmann manifold converges."""
        X = np.array([[1, 3], [4, 1], [2, 1]])
        m0 = 2
        r = 2
        exp = grassmann_manifold_one(X, m0, r)
        self.assertAlmostEqual(exp, 2.60980232)

    def test_F_t(self):
        """Test cost function coverages."""
        X = np.ones((5, 5))
        Y = np.ones((5, 5))
        E = np.zeros((5, 5))
        E[0, 1] = 1
        E[1, 1] = 1
        S = np.eye(5)
        M_E = np.ones((5, 5)) * 6
        M_E[0, 0] = 1
        m0 = 2
        rho = 0.5
        res = cost_function(X, Y, S, M_E, E, m0, rho)
        exp = 1
        assert_array_almost_equal(res, exp, decimal=3)

    def test_F_t_random(self):
        """Test cost function on random values."""
        # random ones and zeros
        np.random.seed(0)
        X = np.ones((5, 5))
        Y = np.ones((5, 5))
        E = np.random.choice([0, 1], size=(5, 5))
        S = np.eye(5)
        M_E = np.ones((5, 5)) * 6
        M_E[0, 0] = 1
        m0 = 2
        rho = 0.5
        res = cost_function(X, Y, S, M_E, E, m0, rho)
        self.assertAlmostEqual(res, 6.5)

    def test_gradF_t(self):
        """Test gradient decent converges."""
        X = np.ones((5, 5))
        Y = np.ones((5, 5))
        E = np.zeros((5, 5))
        E[0, 1] = 1
        E[1, 1] = 1
        S = np.eye(5)
        M_E = np.ones((5, 5)) * 6
        M_E[0, 0] = 1
        m0 = 2
        rho = 0.5
        res = gradient_decent(X, Y, S, M_E, E, m0, rho)
        exp = np.array([[[1., 1., 1., 1., 1.],
                         [1., 1., 1., 1., 1.],
                         [2., 2., 2., 2., 2.],
                         [2., 2., 2., 2., 2.],
                         [2., 2., 2., 2., 2.]],
                        [[2., 2., 2., 2., 2.],
                         [0., 0., 0., 0., 0.],
                         [2., 2., 2., 2., 2.],
                         [2., 2., 2., 2., 2.],
                         [2., 2., 2., 2., 2.]]])
        npt.assert_allclose(exp, res)

    def test_Gp(self):
        """Test second grassmann manifold converges."""
        X = np.ones((5, 5)) * 3
        X[0, 0] = 2
        m0 = 2
        r = 5
        res = grassmann_manifold_two(X, m0, r)
        exp = np.array(
            [[1.08731273, 1.6309691, 1.6309691, 1.6309691, 1.6309691],
             [3.57804989, 3.57804989, 3.57804989, 3.57804989, 3.57804989],
             [3.57804989, 3.57804989, 3.57804989, 3.57804989, 3.57804989],
             [3.57804989, 3.57804989, 3.57804989, 3.57804989, 3.57804989],
             [3.57804989, 3.57804989, 3.57804989, 3.57804989, 3.57804989]]
        )

        npt.assert_allclose(exp, res)

    def test_getoptT(self):
        """Test gradient decent line search."""
        X = np.ones((5, 5))
        Y = np.ones((5, 5))
        E = np.zeros((5, 5))
        E[0, 1] = 1
        E[1, 1] = 1
        S = np.eye(5)
        M_E = np.ones((5, 5)) * 6
        M_E[0, 0] = 1
        m0 = 2
        rho = 0.5
        W, Z = gradient_decent(X, Y, S, M_E, E, m0, rho)
        res = line_search(X, W, Y, Z, S, M_E, E, m0, rho)
        exp = -9.5367431640625e-08
        npt.assert_allclose(exp, res)

    def test_getoptS_small(self):
        """Test singular values from U and V."""
        data = loadmat(get_data_path('small_test.mat'))

        M_E = np.array(data['M_E'].todense())
        E = data['E']

        x = data['x']
        y = data['y']
        res = singular_values(x, y, M_E, E)
        exp = np.array([[0.93639499, 0.07644197, -0.02828782],
                        [-0.03960841, 0.60787383, 0.00521257],
                        [0.00729038, 0.00785834, 0.67853083]])
        npt.assert_allclose(res, exp, atol=1e-5)

    def test_optspace_original(self):
        """Test OptSpace converges on test dataset."""
        M0 = loadmat(get_data_path('large_test.mat'))['M0']
        M_E = loadmat(get_data_path('large_test.mat'))['M_E']

        M0 = M0.astype(np.float)
        M_E = np.array(M_E.todense()).astype(np.float)
        X, S, Y = OptSpace(n_components=3,
                           max_iterations=11,
                           tol=1e-8).solve(M_E)
        err = X[:, ::-1].dot(S).dot(Y[:, ::-1].T) - M0
        n, m = M0.shape

        res = norm(err, 'fro') / np.sqrt(m * n)
        exp = 0.179
        assert_array_almost_equal(res, exp, decimal=1)

    def test_optspace_ordering(self):
        """Test OptSpace produces reproducible loadings."""
        # the expected sorting
        # for U, S, V.
        s_exp = np.array([[5, 4, 1],
                          [8, 3, 0],
                          [7, 9, 2]])
        U_exp = np.array([[6, 3, 0],
                          [7, 4, 1],
                          [8, 5, 2]])
        V_exp = np.array([[6, 3, 0],
                          [7, 4, 1],
                          [8, 5, 2]])
        # un-sorted U,s,v from SVD.
        s_test = np.array([[5, 1, 4],
                           [7, 2, 9],
                           [8, 0, 3]])
        U_test = np.array([[6, 0, 3],
                           [7, 1, 4],
                           [8, 2, 5]])
        V_test = np.array([[6, 0, 3],
                           [7, 1, 4],
                           [8, 2, 5]])
        # run the sorting in optspace
        U_res, s_res, V_res = svd_sort(U_test,
                                       s_test,
                                       V_test)
        assert_array_almost_equal(U_res, U_exp, decimal=3)
        assert_array_almost_equal(s_res, s_exp, decimal=3)
        assert_array_almost_equal(V_res, V_exp, decimal=3)


if __name__ == "__main__":
    unittest.main()



from skbio.util import get_data_path
from qiime2.plugin import ValidationError
from qiime2.plugin.testing import TestPluginBase
from gemelli.q2._format import TrajectoryFormat


class TestTrajectoryFormatTest(TestPluginBase):
    package = "q2_longitudinal.tests"

    def test_valid_simple(self):
        filepath = get_data_path('trajectory.tsv')
        format = TrajectoryFormat(filepath, mode='r')

        format.validate('min')
        format.validate('max')

    def test_valid_real_data(self):
        filepath = get_data_path('context-subject-ordination.tsv')
        format = TrajectoryFormat(filepath, mode='r')

        format.validate('min')
        format.validate('max')

    def test_invalid_header(self):
        filepath = get_data_path('trajectory-invalid-header.tsv')
        format = TrajectoryFormat(filepath, mode='r')
        for level in 'min', 'max':
            with self.assertRaises(ValidationError):
                format.validate(level)

    def test_invalid_non_numeric_column(self):
        filepath = get_data_path('trajectory-non-numeric.tsv')
        format = TrajectoryFormat(filepath, mode='r')
        for level in 'min', 'max':
            with self.assertRaises(ValidationError):
                format.validate(level)



import os
import click
from .__init__ import cli
from biom import load_table
from biom.util import biom_open
from gemelli.preprocessing import (rclr_transformation,
                                   phylogenetic_rclr_transformation)
from gemelli._defaults import DESC_COUNTS, DESC_TREE


@cli.command(name='phylogenetic-rclr')
@click.option('--in-biom',
              help=DESC_COUNTS,
              required=True)
@click.option('--in-phylogeny',
              help=DESC_TREE,
              required=True)
@click.option('--output-dir',
              help='Location of output table.',
              required=True)
def standalone_phylogenetic_rclr(in_biom: str,
                                 in_phylogeny: str,
                                 output_dir: str) -> None:
    """
    Runs phylogenetic robust centered log-ratio transformation.
    Returns both a vectorized table and matched fully labeled phylogeny
    in addition to a rclr transformed version of the phylogenetic table.
    Note: This is run automatically within phylo-CTF/RPCA
    so there no need to run rclr before those functions.
    """

    # import table
    table = load_table(in_biom)
    # run vectorized table and rclr transform
    res_ = phylogenetic_rclr_transformation(table, in_phylogeny)
    counts_by_node, rclr_table, phylogeny = res_

    # If it doesn't already exist, create the output directory.
    # Note that there is technically a race condition here: it's ostensibly
    # possible that some process could delete the output directory after we
    # check that it exists here but before we write the output files to it.
    # However, in this case, we'd just get an error from skbio.io.util.open()
    # (which is called by skbio.OrdinationResults.write()), which makes sense.
    os.makedirs(output_dir, exist_ok=True)

    # write files to output directory
    phylogeny.write(os.path.join(output_dir, 'labeled-phylogeny.nwk'))
    out_path = os.path.join(output_dir, 'phylogenetic-rclr-table.biom')
    with biom_open(out_path, 'w') as wf:
        rclr_table.to_hdf5(wf, "phylogenetic-rclr-table")
    out_path = os.path.join(output_dir, 'phylogenetic-count-table.biom')
    with biom_open(out_path, 'w') as wf:
        counts_by_node.to_hdf5(wf, "phylogenetic-table")


@cli.command(name='rclr')
@click.option('--in-biom',
              help=DESC_COUNTS,
              required=True)
@click.option('--output-dir',
              help='Location of output table.',
              required=True)
def standalone_rclr(in_biom: str,
                    output_dir: str) -> None:
    """
    Runs robust centered log-ratio transformation.
    Note: This is run automatically within CTF/RPCA/Auto-RPCA
    so there no need to run rclr before those functions.
    """

    # import table and perform rclr transform
    table = rclr_transformation(load_table(in_biom))

    # If it doesn't already exist, create the output directory.
    # Note that there is technically a race condition here: it's ostensibly
    # possible that some process could delete the output directory after we
    # check that it exists here but before we write the output files to it.
    # However, in this case, we'd just get an error from skbio.io.util.open()
    # (which is called by skbio.OrdinationResults.write()), which makes sense.
    os.makedirs(output_dir, exist_ok=True)

    # write files to output directory
    out_path = os.path.join(output_dir, 'rclr-table.biom')
    with biom_open(out_path, 'w') as wf:
        table.to_hdf5(wf, "rclr-table")



# ----------------------------------------------------------------------------
# Copyright (c) 2019--, gemelli development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# ----------------------------------------------------------------------------

import warnings
import numpy as np
from numpy.matlib import repmat
from numpy.linalg import norm
from scipy.sparse.linalg import svds


class OptSpace(object):
    """

    OptSpace is a matrix completion algorithm based on a singular value
    decomposition (SVD) optimized on a local manifold. It has been shown to
    be quite robust to noise in low rank datasets (1).
    The objective function that it is trying to optimize
    over is given by:

        min(P|(Y-U*S*V^{T})|_{2}^{2}

    U and V are matrices that are trying to be estimated and S
    is analogous to a matrix of eigenvalues. Y are the
    observed values and P is a function such that
    the errors between Y and USV are only computed
    on the nonzero entries.

    """

    def __init__(
            self,
            n_components,
            max_iterations,
            tol,
            step_size=10000,
            resolution_limit=20,
            sign=-1):
        """

        Parameters
        ----------
        obs: numpy.ndarray - a rclr preprocessed matrix of shape (M,N)
            with missing values set to zero or np.nan.
            N = Features (i.e. OTUs, metabolites)
            M = Samples

        n_components: int or {"optspace"}, optional
            The underlying rank of the dataset.
            This will also control the number of components
            (axis) in the U and V loadings. The value can either
            be hard set by the user or estimated through the
            optspace algorithm.

        max_iterations: int
            The maximum number of convex iterations to optimize the solution
            If iteration is not specified, then the default iteration is 5.
            Which redcues to a satisfactory error threshold.

        tol: float
            Error reduction break, if the error reduced is
            less than this value it will return the solution

        step_size: int, optional : Default is 10000
            The gradient decent step size, this will be
            optimized by the line search.

        resolution_limit: int, optional : Default is 20
            The gradient decent line search resolution limit.
            Where the resolution is line > 2**-resolution_limit.

        sign: int, optional : Default is -1
            Can be one or negative one. This controls the
            sign correction in the gradient decent U, V
            updates.

        Returns
        -------
        self.U: numpy.ndarray - "Sample Loadings" or the unitary matrix
            having left singular vectors as columns.
            Of shape (M, n_components)

        self.s: numpy.ndarray - The singular values,
            sorted in non-increasing order.
            Of shape (n_components, n_components).

        self.V: numpy.ndarray - "Feature Loadings" or Unitary matrix
            having right singular vectors as rows.
            Of shape (N, n_components)

        References
        ----------
        .. [1] Keshavan RH, Oh S, Montanari A. 2009. Matrix completion
                from a few entries (2009_ IEEE International
                Symposium on Information Theory
        """

        self.n_components = n_components
        self.max_iterations = max_iterations
        self.tol = tol
        self.step_size = step_size
        self.resolution_limit = resolution_limit
        self.sign = sign

    def solve(self, obs):
        # adjust iteration indexing by one
        self.max_iterations += 1
        # Convert any nan input to zeros
        # optspace considers zero  and only zero
        # as missing.
        obs[np.isnan(obs)] = 0
        # generate a mask that tracks where missing
        # values exist in the obs dataset
        mask = (np.abs(obs) > 0).astype(np.int)
        # save the shape of the matrix
        n, m = obs.shape
        # get a measure of sparsity level
        total_nonzeros = np.count_nonzero(mask)
        eps = total_nonzeros / np.sqrt(m * n)
        if isinstance(self.n_components, str):
            if self.n_components.lower() == 'auto':
                # estimate the rank of the matrix
                self.n_components = rank_estimate(obs, eps)
                # check estimate again
                if self.n_components >= min(n, m) - 1:
                    warnings.warn('Your matrix is estimated '
                                  'to be high-rank.',
                                  RuntimeWarning)
                # print rank estimate
                print('Estimated rank is %i' % self.n_components)
            else:
                raise ValueError("n-components must be an "
                                 "integer or 'auto'.")
        # raise future warning if hard set
        elif isinstance(self.n_components, int):
            if self.n_components > (min(n, m) - 1):
                raise ValueError("n-components must be at most"
                                 " 1 minus the min. shape of the"
                                 " input matrix.")
        # otherwise rase an error.
        else:
            raise ValueError("n-components must be "
                             "an interger or 'auto'")
        # The rescaling factor compensates the smaller average size of
        # the of the missing entries (mask) with respect to obs
        rescal_param = np.count_nonzero(mask) * self.n_components
        rescal_param = np.sqrt(rescal_param / (norm(obs, 'fro') ** 2))
        obs = obs * rescal_param
        # Our initial first guess are the loadings generated
        # by the traditional SVD
        U, S, V = svds(obs, self.n_components, which='LM')
        # the shape and number of non-zero values
        # can set the input perameters for the gradient
        # decent
        rho = eps * n
        U = U * np.sqrt(n)
        V = (V * np.sqrt(m)).T
        S = S / eps
        # generate the new singular values from
        # the initialization of U and V
        S = singular_values(U, V, obs, mask)
        # initialize the difference between the
        # observed values of the matrix and the
        # the imputed matrix generated by the loadings
        # from this point on we call this "distortion"
        obs_error = obs - U.dot(S).dot(V.T)
        # initialize the distortion matrix of line above
        dist = np.zeros(self.max_iterations + 1)
        # starting initialization of the distortion between obs and imputed
        dist[0] = norm(np.multiply(obs_error, mask), 'fro') / \
            np.sqrt(total_nonzeros)
        # we will perform gradient decent for at most self.max_iterations
        for i in range(1, self.max_iterations):
            # generate new optimized loadings from F(X,Y) [1]
            U_update, V_update = gradient_decent(
                U, V, S, obs, mask, self.step_size, rho)
            # Line search for the optimum jump length in gradient decent
            line = line_search(
                U,
                U_update,
                V,
                V_update,
                S,
                obs,
                mask,
                self.step_size,
                rho,
                resolution_limit=self.resolution_limit)
            # with line and iterations loading update U,V loadings
            U = U - self.sign * line * U_update
            V = V - self.sign * line * V_update
            # generate new singular values from the new
            # loadings
            S = singular_values(U, V, obs, mask)
            # Compute the distortion
            obs_error = obs - U.dot(S).dot(V.T)
            # update the new distortion
            dist[i + 1] = norm(np.multiply(obs_error, mask),
                               'fro') / np.sqrt(total_nonzeros)
            # if the gradient decent has coverged then break the loop
            # and return the results
            if(dist[i + 1] < self.tol):
                break
        # compensates the smaller average size of
        # observed values vs. missing
        S = S / rescal_param
        # ensure the loadings are sorted properly
        U, S, V = svd_sort(U, S, V)
        return U, S, V


def svd_sort(U, S, V):
    """
    Sorting via the s matrix from SVD. In addition to
    sign correction from the U matrix to ensure a
    deterministic output.

    Parameters
    ----------
    U: array-like
        U matrix from SVD
    V: array-like
        V matrix from SVD
    S: array-like
        S matrix from SVD

    Notes
    -----
    S matrix can be off diagonal elements.
    """
    # See https://github.com/scikit-learn/scikit-learn/
    # blob/7b136e92acf49d46251479b75c88cba632de1937/sklearn/
    # decomposition/pca.py#L510-#L518 for context.
    # Because svds do not abide by the normal
    # conventions in scipy.linalg.svd/randomized_svd
    # the output has to be reversed
    idx = np.argsort(np.diag(S))[::-1]
    # sorting following the solution
    # provided by https://stackoverflow.com/
    # questions/36381356/sort-matrix-based
    # -on-its-diagonal-entries
    S = S[idx, :][:, idx]
    U = U[:, idx]
    V = V[:, idx]
    return U, S, V


def cost_function(U, V, S, obs, mask, step_size, rho):
    """
    Parameters
    ----------
    U, V, S, obs, mask, step_size, rho

    Notes
    -----
    M ~ imputed
    """
    # shape of subject loading
    n, n_components = U.shape
    # calculate the Frobenius norm between observed values and imputed values
    distortion = np.sum(
        np.sum((np.multiply((U.dot(S).dot(V.T) - obs), mask)**2))) / 2
    # calculate the  cartesian product of two Grassmann manifolds
    V_manifold = rho * grassmann_manifold_one(V, step_size, n_components)
    U_manifold = rho * grassmann_manifold_one(U, step_size, n_components)
    return distortion + V_manifold + U_manifold


def gradient_decent(U, V, S, obs, mask, step_size, rho):
    """
    A single iteration of the gradient decent update to the U and V matrix.

    Parameters
    ----------
    U, V, S, obs, mask, step_size, rho

    Returns
    -------
    U_update, V_update
    """
    # shape of loadings
    n, n_components = U.shape
    m, n_components = V.shape
    # generate the new values imputed
    # from the loadings from obs values
    US = U.dot(S)
    VS = V.dot(S.T)
    imputed = US.dot(V.T)
    # calculate the U and V distortion
    Qu = U.T.dot(np.multiply((obs - imputed), mask)).dot(VS) / n
    Qv = V.T.dot(np.multiply((obs - imputed), mask).T).dot(US) / m
    # create new loadings based on the distortion between obs and imputed
    # pass these loadings to back to the decent.
    U_update = np.multiply((imputed - obs), mask).dot(VS) + U.dot(Qu) + \
        rho * grassmann_manifold_two(U, step_size, n_components)
    V_update = np.multiply((imputed - obs), mask).T.dot(US) + V.dot(Qv) + \
        rho * grassmann_manifold_two(V, step_size, n_components)
    return U_update, V_update


def line_search(
        U,
        U_update,
        V,
        V_update,
        S,
        obs,
        mask,
        step_size,
        rho,
        resolution_limit=20,
        line=-1e-1):
    """
    An exact line search
        gradient decent converging for a quadratic function

    Parameters
    ----------
    U, U_update, V, V_update, S, obs, mask, step_size, rho
    """

    norm_update = norm(U_update, 'fro')**2 + norm(V_update, 'fro')**2
    # this is the resolution limit (line > 2**-20)
    cost = np.zeros(resolution_limit + 1)
    cost[0] = cost_function(U, V, S, obs, mask, step_size, rho)
    for i in range(resolution_limit):
        cost[i +
             1] = cost_function(U +
                                line *
                                U_update, V +
                                line *
                                V_update, S, obs, mask, step_size, rho)
        if((cost[i + 1] - cost[0]) <= .5 * line * norm_update):
            return line
        line = line / 2
    return line


def singular_values(U, V, obs, mask):
    """
    Generates the singular values from the updated
    U & V loadings for one iteration.

    Parameters
    ----------
    U, V, obs, mask

    """

    n, n_components = U.shape
    C = np.ravel(U.T.dot(obs).dot(V))
    A = np.zeros((n_components * n_components, n_components * n_components))
    for i in range(n_components):
        for j in range(n_components):
            ind = j * n_components + i
            x = U[:, i].reshape(1, len(U[:, i]))
            manifold = V[:, j].reshape(len(V[:, j]), 1)
            tmp = np.multiply((manifold.dot(x)).T, mask)
            temp = U.T.dot(tmp).dot(V)
            A[:, ind] = np.ravel(temp)
    S = np.linalg.lstsq(A, C, rcond=1e-12)[0]
    S = S.reshape((n_components, n_components)).T

    return S


def grassmann_manifold_one(U, step_size, n_components):
    """
    The first Grassmann Manifold

    Parameters
    ----------
    U, step_size, n_components
    """
    # get the step from the manifold
    step = np.sum(U**2, axis=1) / (2 * step_size * n_components)
    manifold = np.exp((step - 1)**2) - 1
    manifold[step < 1] = 0
    manifold[manifold == np.inf] = 0
    return manifold.sum()


def grassmann_manifold_two(U, step_size, n_components):
    """
    The second Grassmann Manifold

    Parameters
    ----------
    U, step_size, n_components


    """
    # get the step from the manifold
    step = np.sum(U**2, axis=1) / (2 * step_size * n_components)
    step = 2 * np.multiply(np.exp((step - 1)**2), (step - 1))
    step[step < 0] = 0
    step = step.reshape(len(step), 1)
    step = np.multiply(U, repmat(step, 1, n_components)) / \
        (step_size * n_components)
    return step


def rank_estimate(obs, eps, k=20, lam=0.05,
                  min_rank=3, max_iter=5000):
    """
    This function estimates the rank of a
    sparse matrix (i.e. with missing values).

    Parameters
    ----------
    obs: numpy.ndarray - a rclr preprocessed matrix of shape (M,N)
        with missing values set to zero or np.nan.
        N = Features (i.e. OTUs, metabolites)
        M = Samples

    eps: float - Measure of the level of sparsity
        Equivalent to obs N-non-zeros / sqrt(obs.shape)

    k: int - Max number of singular values / rank

    lam: float - Step in the iteration

    min_rank: int - The min. rank allowed

    Returns
    -------
    The estimated rank of the matrix.

    References
    ----------
    .. [1] Part C in Keshavan, R. H., Montanari,
           A. & Oh, S. Low-rank matrix completion
           with noisy observations: A quantitative
           comparison. in 2009 47th Annual Allerton
           Conference on Communication, Control,
           and Computing (Allerton) 1216–1222 (2009).
    """
    # dim. of the data
    n, m = obs.shape
    # ensure rank worth estimating
    if min(n, m) <= 2:
        return min_rank
    # get N-singular values
    s = svds(obs,  min(k, n, m) - 1, which='LM',
             return_singular_vectors=False)[::-1]
    # get N+1 singular values
    s_one = s[:-1] - s[1:]
    # simplify iterations
    s_one_ = s_one / np.mean(s_one[-10:])
    # iteration one
    r_one = 0
    iter_ = 0
    while r_one <= 0:
        cost = []
        # get the cost
        for idx in range(s_one_.shape[0]):
            cost.append(lam * max(s_one_[idx:]) + idx)
        # estimate the rank
        r_one = np.argmin(cost)
        lam += lam
        iter_ += 1
        if iter_ > max_iter:
            break

    # iteration two
    cost = []
    # estimate the rank
    for idx in range(s.shape[0] - 1):
        cost.append((s[idx + 1] + np.sqrt(idx * eps) * s[0] / eps) / s[idx])
    r_two = np.argmin(cost)
    # return the final estimate
    return max(r_one, r_two, min_rank)



import unittest
import os
import numpy as np
from skbio import TreeNode
from qiime2 import Artifact
import numpy.testing as npt
from biom import load_table, Table
from biom.util import biom_open
from os.path import sep as os_path_sep
from click.testing import CliRunner
from skbio.util import get_data_path
from numpy.testing import assert_array_almost_equal
from qiime2.plugins.gemelli.actions import (rclr_transformation,
                                            phylogenetic_rclr_transformation)
from gemelli.scripts.__init__ import cli as sdc


class Test_qiime2_rclr(unittest.TestCase):

    def setUp(self):
        self.cdata = np.array([[3, 3, 0],
                               [0, 4, 2]])
        self.true = np.array([[0.0, 0.0, np.nan],
                              [np.nan, 0.34657359, -0.34657359]])
        pass

    def test_qiime2_rclr(self):
        """Tests q2-rclr matches standalone rclr."""

        # make mock table to write
        samps_ids = ['s%i' % i for i in range(self.cdata.shape[0])]
        feats_ids = ['f%i' % i for i in range(self.cdata.shape[1])]
        table_test = Table(self.cdata.T, feats_ids, samps_ids)
        # write table
        in_ = get_data_path('test-smallest.biom',
                            subfolder='data')
        out_path = os_path_sep.join(in_.split(os_path_sep)[:-1])
        test_path = os.path.join(out_path, 'rclr-test.biom')
        with biom_open(test_path, 'w') as wf:
            table_test.to_hdf5(wf, "test")
        # run standalone
        runner = CliRunner()
        result = runner.invoke(sdc.commands['rclr'],
                               ['--in-biom', test_path,
                                '--output-dir', out_path])
        out_table = get_data_path('rclr-table.biom',
                                  subfolder='data')
        res_table = load_table(out_table)
        standalone_mat = res_table.matrix_data.toarray().T
        # check that exit code was 0 (indicating success)
        try:
            self.assertEqual(0, result.exit_code)
        except AssertionError:
            ex = result.exception
            error = Exception('Command failed with non-zero exit code')
            raise error.with_traceback(ex.__traceback__)
        # run QIIME2
        q2_table_test = Artifact.import_data("FeatureTable[Frequency]",
                                             table_test)
        q2_res = rclr_transformation(q2_table_test).rclr_table.view(Table)
        q2_res_mat = q2_res.matrix_data.toarray().T
        # check same and check both correct
        npt.assert_allclose(standalone_mat, q2_res_mat)
        npt.assert_allclose(standalone_mat, self.true)
        npt.assert_allclose(q2_res_mat, self.true)

    def test_qiime2_phylogenetic_rclr(self):
        """Tests q2-phylogenetic-rclr matches standalone rclr."""

        # write table
        in_table = get_data_path('test.biom')
        in_tree = get_data_path('tree.nwk')
        out_path = os_path_sep.join(in_table.split(os_path_sep)[:-1])

        # run phylogenetic-rclr
        runner = CliRunner()
        result = runner.invoke(sdc.commands['phylogenetic-rclr'],
                               ['--in-biom', in_table,
                                '--in-phylogeny', in_tree,
                                '--output-dir', out_path])
        # check that exit code was 0 (indicating success)
        try:
            self.assertEqual(0, result.exit_code)
        except AssertionError:
            ex = result.exception
            error = Exception('Command failed with non-zero exit code')
            raise error.with_traceback(ex.__traceback__)

        # import result
        tree_res = get_data_path('labeled-phylogeny.nwk')
        tree_res = TreeNode.read(tree_res, format='newick')
        counts_res = get_data_path('phylogenetic-count-table.biom')
        counts_res = load_table(counts_res)
        rclr_res = get_data_path('phylogenetic-rclr-table.biom')
        rclr_res = load_table(rclr_res)

        # run QIIME2
        table_test = load_table(in_table)
        q2_table_test = Artifact.import_data("FeatureTable[Frequency]",
                                             table_test)
        tree_test = TreeNode.read(in_tree,
                                  format='newick')
        q2_tree_test = Artifact.import_data("Phylogeny[Rooted]",
                                            tree_test)
        q2_res = phylogenetic_rclr_transformation(q2_table_test,
                                                  q2_tree_test)
        cbn_table, cbn_rclr_table, cbn_tree = q2_res
        q2tree = cbn_tree.view(TreeNode)
        q2cbn_table = cbn_table.view(Table)
        q2cbn_rclr_table = cbn_rclr_table.view(Table)

        # check table values match
        assert_array_almost_equal(q2cbn_table.matrix_data.toarray(),
                                  counts_res.matrix_data.toarray())
        assert_array_almost_equal(q2cbn_rclr_table.matrix_data.toarray(),
                                  rclr_res.matrix_data.toarray())

        # check renamed names are consistent
        name_check_ = [x.name == y.name for x, y in zip(tree_res.postorder(),
                                                        q2tree.postorder())]
        name_check_ = all(name_check_)
        self.assertEqual(name_check_, True)



import unittest
import numpy as np
import pandas as pd
from pandas import read_csv
from qiime2 import Artifact
from click.testing import CliRunner
from nose.tools import nottest
from biom import Table, load_table
from skbio.util import get_data_path
from os.path import sep as os_path_sep
from gemelli.rpca import rpca, auto_rpca
from gemelli.scripts.__init__ import cli as sdc
from gemelli.simulations import build_block_model
from qiime2.plugins import gemelli as q2gemelli
from skbio import OrdinationResults, TreeNode
from skbio.stats.distance import DistanceMatrix
from numpy.testing import assert_array_almost_equal
from gemelli.testing import assert_ordinationresults_equal


@nottest
def create_test_table():
    _, test_table = build_block_model(rank=2,
                                      hoced=20,
                                      hsced=20,
                                      spar=2e3,
                                      C_=2e3,
                                      num_samples=50,
                                      num_features=500,
                                      mapping_on=False)

    feat_ids = ['F%d' % i for i in range(test_table.shape[0])]
    samp_ids = ['L%d' % i for i in range(test_table.shape[1])]

    return Table(test_table, feat_ids, samp_ids)


class Testrpca(unittest.TestCase):

    def setUp(self):
        self.test_table = create_test_table()

    def test_rpca(self):
        """Tests the basic validity of the actual rpca() method's outputs."""
        ord_test, dist_test = rpca(table=self.test_table)
        # Validate types of the RPCA outputs
        self.assertIsInstance(ord_test, OrdinationResults)
        self.assertIsInstance(dist_test, DistanceMatrix)
        # Ensure that no NaNs are in the OrdinationResults
        # NOTE that we have to use the DataFrame .any() functions instead of
        # python's built-in any() functions -- see #29 for details on this
        self.assertFalse(np.isnan(ord_test.features).any(axis=None))
        self.assertFalse(np.isnan(ord_test.samples).any(axis=None))

    def test_auto_rpca(self):
        """Tests the basic validity of the actual auto_rpca()."""
        ord_test, dist_test = auto_rpca(table=self.test_table)
        # Validate types of the RPCA outputs
        self.assertIsInstance(ord_test, OrdinationResults)
        self.assertIsInstance(dist_test, DistanceMatrix)
        # Ensure that no NaNs are in the OrdinationResults
        # NOTE that we have to use the DataFrame .any() functions instead of
        # python's built-in any() functions -- see #29 for details on this
        self.assertFalse(np.isnan(ord_test.features).any(axis=None))
        self.assertFalse(np.isnan(ord_test.samples).any(axis=None))


class Test_qiime2_rpca(unittest.TestCase):

    def setUp(self):
        self.q2table = Artifact.import_data("FeatureTable[Frequency]",
                                            create_test_table())

    def test_qiime2_auto_rpca(self):
        """ Test Q2 rank estimate matches standalone."""

        tstdir = "test_output"
        # Run gemelli through QIIME 2 (specifically, the Artifact API)
        res = q2gemelli.actions.auto_rpca(self.q2table)
        ordination_qza, distmatrix_qza = res
        # Get the underlying data from these artifacts
        # q2ordination = ordination_qza.view(OrdinationResults)
        q2distmatrix = distmatrix_qza.view(DistanceMatrix)

        # Next, run gemelli outside of QIIME 2. We're gonna check that
        # everything matches up.
        # ...First, though, we need to write the contents of self.q2table to a
        # BIOM file, so gemelli can understand it.
        self.q2table.export_data(get_data_path("", tstdir))
        q2table_loc = get_data_path('feature-table.biom', tstdir)
        # Derived from a line in test_standalone_rpca()
        tstdir_absolute = os_path_sep.join(q2table_loc.split(os_path_sep)[:-1])

        # Run gemelli outside of QIIME 2...
        result = CliRunner().invoke(sdc.commands['auto-rpca'],
                                    ['--in-biom', q2table_loc,
                                     '--output-dir', tstdir_absolute])
        # ...and read in the resulting output files. This code was derived from
        # test_standalone_rpca() elsewhere in gemelli's codebase.
        # stordination = OrdinationResults.read(get_data_path('ordination.txt',
        #                                                    tstdir))
        stdistmatrix_values = read_csv(
            get_data_path(
                'distance-matrix.tsv',
                tstdir),
            sep='\t',
            index_col=0).values

        # Convert the DistanceMatrix object a numpy array (which we can compare
        # with the other _values numpy arrays we've created from the other
        # distance matrices)
        q2distmatrix_values = q2distmatrix.to_data_frame().values

        # Finaly: actually check the consistency of Q2 and standalone results!
        np.testing.assert_array_almost_equal(q2distmatrix_values,
                                             stdistmatrix_values)
        # check that exit code was 0 (indicating success)
        try:
            self.assertEqual(0, result.exit_code)
        except AssertionError:
            ex = result.exception
            error = Exception('Command failed with non-zero exit code')
            raise error.with_traceback(ex.__traceback__)

    def test_qiime2_rpca(self):
        """Tests that the Q2 and standalone RPCA results match."""

        tstdir = "test_output"
        # Run gemelli through QIIME 2 (specifically, the Artifact API)
        ordination_qza, distmatrix_qza = q2gemelli.actions.rpca(self.q2table)
        # Get the underlying data from these artifacts
        # q2ordination = ordination_qza.view(OrdinationResults)
        q2distmatrix = distmatrix_qza.view(DistanceMatrix)

        # Next, run gemelli outside of QIIME 2. We're gonna check that
        # everything matches up.
        # ...First, though, we need to write the contents of self.q2table to a
        # BIOM file, so gemelli can understand it.
        self.q2table.export_data(get_data_path("", tstdir))
        q2table_loc = get_data_path('feature-table.biom', tstdir)
        # Derived from a line in test_standalone_rpca()
        tstdir_absolute = os_path_sep.join(q2table_loc.split(os_path_sep)[:-1])

        # Run gemelli outside of QIIME 2...
        result = CliRunner().invoke(sdc.commands['rpca'],
                                    ['--in-biom', q2table_loc,
                                     '--output-dir', tstdir_absolute])
        # ...and read in the resulting output files. This code was derived from
        # test_standalone_rpca() elsewhere in gemelli's codebase.
        # stordination = OrdinationResults.read(get_data_path('ordination.txt',
        #                                                    tstdir))
        stdistmatrix_values = read_csv(
            get_data_path(
                'distance-matrix.tsv',
                tstdir),
            sep='\t',
            index_col=0).values

        # Convert the DistanceMatrix object a numpy array (which we can compare
        # with the other _values numpy arrays we've created from the other
        # distance matrices)
        q2distmatrix_values = q2distmatrix.to_data_frame().values

        # Finaly: actually check the consistency of Q2 and standalone results!
        np.testing.assert_array_almost_equal(q2distmatrix_values,
                                             stdistmatrix_values)
        # check that exit code was 0 (indicating success)
        try:
            self.assertEqual(0, result.exit_code)
        except AssertionError:
            ex = result.exception
            error = Exception('Command failed with non-zero exit code')
            raise error.with_traceback(ex.__traceback__)

        # NOTE: This functionality is currently not used due to the inherent
        # randomness in how the test table data is generated (and also because
        # we're already checking the correctness of the standalone gemelli
        # RPCA script), but if desired you can add ground truth data to a data/
        # folder in this directory (i.e. a distance-matrix.tsv and
        # ordination.txt file), and the code below will compare the Q2 results
        # to those files.
        #
        # Read in expected output from data/, similarly to above
        # exordination = OrdinationResults.read(get_data_path(
        #                                       'ordination.txt'))
        # exdistmatrix_values = read_csv(get_data_path('distance-matrix.tsv'),
        #                                sep='\t', index_col=0).values
        #
        # ... And check consistency of Q2 results with the expected results
        # assert_gemelli_ordinationresults_equal(q2ordination, exordination)
        # np.testing.assert_array_almost_equal(q2distmatrix_values,
        #                                      exdistmatrix_values)

    def test_qiime2_phylogenetic_rpca(self):
        """Tests that the Q2 (without taxonomy) & standalone phylogenetic RPCA
           match.
        """

        in_table = get_data_path('test.biom')
        in_tree = get_data_path('tree.nwk')
        out_ = os_path_sep.join(in_table.split(os_path_sep)[:-1])
        runner = CliRunner()
        result = runner.invoke(sdc.commands['phylogenetic-rpca'],
                               ['--in-biom', in_table,
                                '--in-phylogeny', in_tree,
                                '--output-dir', out_])
        # Read the results
        dist_res = pd.read_csv(get_data_path('distance-matrix.tsv'),
                               sep='\t',
                               index_col=0)
        ord_res = OrdinationResults.read(get_data_path('ordination.txt'))
        tree_res = TreeNode.read(get_data_path('labeled-phylogeny.nwk'),
                                 format='newick')
        bt_res = load_table(get_data_path('phylo-table.biom'))

        # Run gemelli through QIIME 2 (specifically, the Artifact API)
        table_test = load_table(in_table)
        q2_table_test = Artifact.import_data("FeatureTable[Frequency]",
                                             table_test)
        tree_test = TreeNode.read(in_tree,
                                  format='newick')
        q2_tree_test = Artifact.import_data("Phylogeny[Rooted]",
                                            tree_test)
        res = q2gemelli.actions.phylogenetic_rpca_without_taxonomy(
            q2_table_test,
            q2_tree_test)
        # biplot, distance, count-tree, count-table
        q2ord, q2dist, q2ctree, q2ctbl = res
        # Get the underlying data from these artifacts
        q2ord = q2ord.view(OrdinationResults)
        q2dist = q2dist.view(DistanceMatrix)
        q2dist = q2dist.to_data_frame()
        q2ctree = q2ctree.view(TreeNode)
        q2ctbl = q2ctbl.view(Table)

        # check table values match
        assert_array_almost_equal(bt_res.matrix_data.toarray(),
                                  q2ctbl.matrix_data.toarray())

        # check renamed names are consistent
        name_check_ = [x.name == y.name for x, y in zip(tree_res.postorder(),
                                                        q2ctree.postorder())]
        name_check_ = all(name_check_)
        self.assertEqual(name_check_, True)

        # Check that the distance matrix matches our expectations
        assert_array_almost_equal(dist_res.values, q2dist.values)

        # Check that the ordination results match our expectations -- checking
        # each value for both features and samples
        assert_ordinationresults_equal(ord_res, q2ord)

        # check that exit code was 0 (indicating success)
        try:
            self.assertEqual(0, result.exit_code)
        except AssertionError:
            ex = result.exception
            error = Exception('Command failed with non-zero exit code')
            raise error.with_traceback(ex.__traceback__)


if __name__ == "__main__":
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2019--, gemelli development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# ----------------------------------------------------------------------------

import warnings
import io
import sys
import os
import t2t.nlevel as nl
import numpy as np
import pandas as pd
from biom import Table
from skbio import TreeNode
from .base import _BaseConstruct
from gemelli._defaults import DEFAULT_MTD
from skbio.diversity._util import _vectorize_counts_and_tree
from bp import parse_newick, to_skbio_treenode
# import QIIME2 if in a Q2env otherwise set type to str
try:
    from q2_types.tree import NewickFormat
except ImportError:
    # python does not check but technically this is the type
    NewickFormat = str


VALID_TAXONOMY_COLUMN_NAMES = ('taxon', 'taxonomy')


class TaxonomyError(Exception):
    pass


def create_taxonomy_metadata(phylogeny, traversed_taxonomy=None):
    """
        Create a pd.DataFrame with index 'Feature ID' and column 'Taxon'.

        This method will traverse phylogeny using TreeNode.traverse() and
        use the node names found in phylogeny as the index 'Feature ID'.
        Thus, traversed_taxonomy must be parallel List of phylogeny (i.e. a
        List that layes out the taxonomy in a traverse() fashion).

        Parameters
        ----------
        phylogeny: TreeNode
        traversed_taxonomy: List
    """
    f_id = []
    tax = []
    if traversed_taxonomy is not None:
        # add taxonomy for all nodes in tree
        for i, node in enumerate(phylogeny.traverse(include_self=True)):
            # metadata.append([node.name, traversed_taxonomy[i]])
            f_id.append(node.name)
            tax.append(traversed_taxonomy[i])
    else:
        # create empty dataframe
        f_id = ['None']
        tax = ['None']
    returned_taxonomy = pd.DataFrame(data={'Feature ID': f_id, 'Taxon': tax})
    returned_taxonomy.set_index(keys='Feature ID', inplace=True)
    return returned_taxonomy


def retrieve_t2t_taxonomy(phylogeny, taxonomy=None):
    """
    Returns a List containing the taxonomy of all nodes in the tree stored in
    TreeNode.traverse() order. If taxonomy is None, then None will be returned

    based on :
    https://github.com/biocore/tax2tree/blob/9b3814fb19e935c06a31e61e848d0f91bcecb305/scripts/t2t#L46

    Parameters
    ----------
    phylogeny: TreeNode
    taxonomy : pd.DataFrame with Index 'Feature ID' and contains a column
        'taxon' or 'taxonomy' (case insensitive)
    """
    if taxonomy is None:
        # return empty taxonomy
        return None

    # make copy of phylogeny
    consensus_tree = phylogeny.copy()
    # validate and convert taxonomy into a StringIO stream
    consensus_map = _get_taxonomy_io_stream(taxonomy)

    tipname_map = nl.load_consensus_map(consensus_map, False)
    tree_ = nl.load_tree(consensus_tree, tipname_map)
    counts = nl.collect_names_at_ranks_counts(tree_)

    nl.decorate_ntips(tree_)
    nl.decorate_name_relative_freqs(tree_, counts, 2)
    nl.set_ranksafe(tree_)
    nl.pick_names(tree_)

    nl.set_preliminary_name_and_rank(tree_)

    contree, contree_lookup = nl.make_consensus_tree(tipname_map.values())
    # Disable print statements
    with open(os.devnull, 'w') as fpnull:
        sys.stdout = fpnull
    sys.stdout = open(os.devnull, 'w')
    nl.backfill_names_gap(tree_, contree_lookup)
    # Restore print statements
    sys.stdout = sys.__stdout__
    nl.commonname_promotion(tree_)

    constrings = _pull_consensus_strings(tree_)
    return constrings


def _pull_consensus_strings(consensus_tree):
    """
    Pulls consensus strings off of consensus_tree. Assumes .name is set
    This is a helper function for retrieve_t2t_taxonomy.

    based on:
    https://github.com/biocore/tax2tree/blob/9b3814fb19e935c06a31e61e848d0f91bcecb305/t2t/nlevel.py#L831

    Parameters
    ----------
    consensus_tree: TreeNode
    """
    constrings = []
    rank_order = {r: i for i, r in enumerate(nl.RANK_ORDER)}

    # helper function
    def _add_name_to_consensus_string(node_name, cons_string):
        # only add if node has a name
        if node_name:
            if ';' in node_name:
                names = [r.strip() for r in node_name.split(';')]
                for node_name in names:
                    rank_idx = rank_order[node_name[0]]
                    cons_string[rank_idx] = node_name
            else:
                rank_idx = rank_order[node_name[0]]
                cons_string[rank_idx] = node_name

    # start at each node and travel up
    for node in consensus_tree.traverse(include_self=True):
        consensus_string = ['%s__' % r for r in nl.RANK_ORDER]
        # internal nodes will have taxonomy for name so we need to add it to
        # the consensus_string
        if not node.is_tip():
            _add_name_to_consensus_string(node.name, consensus_string)
        if node.is_root():
            constrings.append('; '.join(consensus_string))
            continue
        p = node.parent
        # walk up the consensus_tree filling in the consensus string
        while p.parent:
            _add_name_to_consensus_string(p.name, consensus_string)
            p = p.parent

        # if there is a name at the root we need to make sure we grab it
        if p.name:
            _add_name_to_consensus_string(p.name, consensus_string)

        # join strings with tip id
        constrings.append('; '.join(consensus_string))
    return constrings


def _get_taxonomy_io_stream(taxonomy):
    """
    Returns a StringIO of taxonomy.
    This is a hepler function for retrieve_t2t_taxonomy.
    Raises
    ------
    TaxonomyError
        taxonomy does not contain a valid taxonomy column (see
        VALID_TAXONOMY_COLUMN_NAMES )
    """
    lowercase_col_names = [str(c).lower() for c in taxonomy.columns]

    # See if there is a "taxonomy column", and do some related validation on
    # column names
    tax_col_index = None
    tax_col_name = None
    for i, col in enumerate(lowercase_col_names):
        if col in VALID_TAXONOMY_COLUMN_NAMES:
            if tax_col_index is None:
                tax_col_index = i
                # ("col" has already been set to lowercase)
                tax_col_name = taxonomy.columns[i]
            else:
                # Error condition 1 -- multiple possible "taxonomy columns" :(
                raise TaxonomyError(
                    (
                        "Multiple columns in the feature metadata have one of "
                        "the following names (case insensitive): {}. At most "
                        "one feature metadata column can have a name from "
                        "that list."
                    ).format(VALID_TAXONOMY_COLUMN_NAMES)
                )
    if tax_col_name is None:
        raise TaxonomyError(
            (
                "The taxonomy file does not contain a column with one the "
                "following (case insensitive): {}."
            ).format(VALID_TAXONOMY_COLUMN_NAMES)
        )
    stream = io.StringIO()

    # Split the single column of taxonomy strings into n columns, where n
    # is the highest number of taxonomic levels in any string. This is to
    # ensure that all feateures have the same amount of levels for t2t
    taxonomy = taxonomy[tax_col_name]\
        .str.strip().str.split(r'\s*;\s*', expand=True)

    #  set t2t rank order
    highest_rank_indx = taxonomy.dropna(axis=0, how='any').first_valid_index()
    rank_order = [rank[0]
                  for rank in taxonomy.loc[highest_rank_indx].values.tolist()]
    nl.set_rank_order(rank_order)

    # collapse taxonomy back into a single string
    taxonomy.fillna('', inplace=True)
    taxonomy = taxonomy.apply(
        lambda levels: '; '.join(levels.values.tolist()),
        axis=1)

    # convert taxonomy dataframe to a StringIO for use in t2t
    taxonomy.to_csv(stream, sep='\t', index=True, header=False)

    # set stream to point to first line
    stream.seek(0)

    return stream


def bp_read_phylogeny(table: Table,
                      phylogeny: NewickFormat,
                      min_depth: int = DEFAULT_MTD):
    """
    Fast way to read in phylogeny in newick
    format, filter, and return in TreeNode format.

    Parameters
    ----------
    table: biom.Table - a table of shape (M,N)
        N = Features (i.e. OTUs, metabolites)
        M = Samples
    phylogeny: str - path to file/data
                     in newick format
    min_depth: int
        Minimum number of total number of
        descendants (tips) to include a node.
        Default value of zero will retain all nodes
        (including tips).
    Examples
    --------
    TODO

    """

    # import file path
    with open(str(phylogeny)) as treefile:
        # read balanced parentheses tree
        phylogeny = parse_newick(treefile.readline())
        # first filter out
        names_to_keep = set((table.ids('observation')).flatten())
        phylogeny = phylogeny.shear(names_to_keep).collapse()
        # convert the tree to skbio TreeNode for processing
        phylogeny = to_skbio_treenode(phylogeny)
        # filter internal nodes based on topology
        tree_topology_filter(phylogeny, min_depth=min_depth)

    return phylogeny


def tensor_rclr(T, branch_lengths=None):
    """
    Robust clr transform. is the approximate geometric mean of X.

    We know from the Central Limit Theorem that as
    we collect more independent measurements we
    approach the true geometric mean.

    This transformation will work on N mode tensors
    by flattening. Flattened tensor are reshaped into
    subject x contions by features before transformation.
    A tensor will be returned in the shape shape and order.

    Mode 2 tensors (matrix) will be directly transformed,
    no reshaping necessary.

    Parameters
    ----------
    T : array-like
        Array of non-negative count data.
        In an N mode tensor of shape:
        first dimension = samples
        second dimension = features
        [3..N] dimensions = conditions

    Raises
    ------
    ValueError
        Tensor is less than 2-dimensions.
    ValueError
        Tensor contains negative values.
    ValueError
        Tensor contains np.inf or -np.inf.
    ValueError
        Tensor contains np.nan or missing.

    References
    ----------
    .. [1] V. Pawlowsky-Glahn, J. J. Egozcue,
           R. Tolosana-Delgado (2015),
           Modeling and Analysis of
           Compositional Data, Wiley,
           Chichester, UK

    .. [2] C. Martino et al., A Novel Sparse
           Compositional Technique Reveals
           Microbial Perturbations. mSystems.
           4 (2019), doi:10.1128/mSystems.00016-19.

    Examples
    --------
    TODO

    """

    if len(T.shape) < 2:
        raise ValueError('Tensor is less than 2-dimensions')

    if np.count_nonzero(np.isinf(T)) != 0:
        raise ValueError('Tensor contains either np.inf or -np.inf.')

    if np.count_nonzero(np.isnan(T)) != 0:
        raise ValueError('Tensor contains np.nan or missing.')

    if (T < 0).any():
        raise ValueError('Tensor contains negative values.')

    if len(T.shape) < 3:
        # tensor_rclr on 2D matrix
        M_tensor_rclr = matrix_rclr(T.transpose().copy(),
                                    branch_lengths=branch_lengths).T
        M_tensor_rclr[~np.isfinite(M_tensor_rclr)] = 0.0
        return M_tensor_rclr
    else:
        # flatten tensor (samples*conditions x features)
        T = T.copy()
        # conditional dimensions
        conditions_index = list(range(2, len(T.shape)))
        forward_T = tuple([0] + conditions_index + [1])
        reverse_T = tuple([0] + [conditions_index[-1]]
                          + [1] + conditions_index[:-1])
        # transpose to flatten
        T = T.transpose(forward_T)
        M = T.reshape(np.product(T.shape[:len(T.shape) - 1]),
                      T.shape[-1])
        with np.errstate(divide='ignore', invalid='ignore'):
            M_tensor_rclr = matrix_rclr(M, branch_lengths=branch_lengths)
        M_tensor_rclr[~np.isfinite(M_tensor_rclr)] = 0.0
        # reshape to former tensor and return tensors
        return M_tensor_rclr.reshape(T.shape).transpose(reverse_T)


def matrix_rclr(mat, branch_lengths=None):
    """
    Robust clr transform helper function.
    This function is built for mode 2 tensors,
    also known as matrices.

    Raises
    ------
    ValueError
       Raises an error if any values are negative.
    ValueError
       Raises an error if the matrix has more than 2 dimension.

    References
    ----------
    .. [1] V. Pawlowsky-Glahn, J. J. Egozcue,
           R. Tolosana-Delgado (2015),
           Modeling and Analysis of
           Compositional Data, Wiley,
           Chichester, UK

    .. [2] C. Martino et al., A Novel Sparse
           Compositional Technique Reveals
           Microbial Perturbations. mSystems.
           4 (2019), doi:10.1128/mSystems.00016-19.

    Examples
    --------
    TODO

    """
    # ensure array is at least 2D
    mat = np.atleast_2d(np.array(mat))
    # ensure array not more than 2D
    if mat.ndim > 2:
        raise ValueError("Input matrix can only have two dimensions or less")
    # ensure no neg values
    if (mat < 0).any():
        raise ValueError('Array Contains Negative Values')
    # ensure no undefined values
    if np.count_nonzero(np.isinf(mat)) != 0:
        raise ValueError('Data-matrix contains either np.inf or -np.inf')
    # ensure no missing values
    if np.count_nonzero(np.isnan(mat)) != 0:
        raise ValueError('Data-matrix contains nans')
    # take the log of the sample centered data
    if branch_lengths is not None:
        mat = np.log(matrix_closure(matrix_closure(mat) * branch_lengths))
    else:
        mat = np.log(matrix_closure(mat))
    # generate a mask of missing values
    mask = [True] * mat.shape[0] * mat.shape[1]
    mask = np.array(mat).reshape(mat.shape)
    mask[np.isfinite(mat)] = False
    # sum of rows (features)
    lmat = np.ma.array(mat, mask=mask)
    # perfrom geometric mean
    gm = lmat.mean(axis=-1, keepdims=True)
    # center with the geometric mean
    lmat = (lmat - gm).squeeze().data
    # mask the missing with nan
    lmat[~np.isfinite(mat)] = np.nan
    return lmat


def rclr_transformation(table: Table) -> Table:
    """
    Takes biom table and returns
    a matrix_rclr transformed biom table.
    """
    # transform table values (and return biom.Table)
    table = Table(matrix_rclr(table.matrix_data.toarray().T).T,
                  table.ids('observation'),
                  table.ids('sample'))
    return table


def phylogenetic_rclr_transformation(table: Table,
                                     phylogeny: NewickFormat,
                                     min_depth: int = DEFAULT_MTD) -> (
                                         Table, Table, TreeNode):
    """
    Takes biom table and returns fast_unifrac style
    vectorized count table and a matrix_rclr
    transformed biom table.

    """

    # import the tree and filter
    phylogeny = bp_read_phylogeny(table,
                                  phylogeny,
                                  min_depth)
    # build the vectorized table
    counts_by_node, tree_index, branch_lengths, fids, otu_ids\
        = fast_unifrac(table, phylogeny)
    # Robust-clt (matrix_rclr) preprocessing
    rclr_table = matrix_rclr(counts_by_node, branch_lengths=branch_lengths)
    # import transformed matrix into biom.Table
    rclr_table = Table(rclr_table.T,
                       fids, table.ids('sample'))
    # import expanded matrix into biom.Table
    counts_by_node = Table(counts_by_node.T,
                           fids, table.ids())

    return counts_by_node, rclr_table, phylogeny


def matrix_closure(mat):
    """
    Simillar to the skbio.stats.composition.closure function.
    Performs closure to ensure that all elements add up to 1.
    However, this function allows for zero rows. This results
    in rows that may contain missing (NaN) vlues. These
    all zero rows may occur as a product of a tensor slice and
    is dealt later with the tensor restructuring and factorization.

    Parameters
    ----------
    mat : array_like
       a matrix of proportions where
       rows = compositions
       columns = components
    Returns
    -------
    array_like, np.float64
       A matrix of proportions where all of the values
       are nonzero and each composition (row) adds up to 1
    Examples
    --------
    >>> import numpy as np
    >>> from gemelli.preprocessing import matrix_closure
    >>> X = np.array([[2, 2, 6], [0, 0, 0]])
    >>> closure(X)
    array([[ 0.2,  0.2,  0.6],
           [ nan,  nan,  nan]])

    """

    mat = np.atleast_2d(mat)
    mat = mat / mat.sum(axis=1, keepdims=True)

    return mat.squeeze()


def fast_unifrac(table, tree):
    """
    A wrapper to return a vectorized Fast UniFrac
    algorithm. The nodes up the tree are summed
    and exposed as vectors in the matrix. The
    closed matrix is then multipled by the
    branch lengths to phylogenically
    weight the data.

    Parameters
    ----------
    table : biom.Table
       A biom table of counts.
    tree: skbio.TreeNode
       Tree containing the features in the table.
    Returns
    -------
    counts_by_node: array_like, np.float64
       A matrix of counts with internal nodes
       vectorized.
    tree_index: dict
        A housekeeping dictionary.
    branch_lengths: array_like, np.float64
        An array of branch lengths.
    fids: list
        A list of feature IDs matched to tree_index['id'].
    otu_ids: list
        A list of the original table OTU IDs (tips).
    Examples
    --------
    TODO

    """

    # original table
    bt_array = table.matrix_data.toarray()
    otu_ids = table.ids('observation')
    # expand the vectorized table
    counts_by_node, tree_index, branch_lengths \
        = _vectorize_counts_and_tree(bt_array.T, otu_ids, tree)
    # check branch lengths
    if sum(branch_lengths) == 0:
        raise ValueError('All tree branch lengths are zero. '
                         'This will result in a table of zero features.')
    # drop zero sum features (non-optional for CTF/RPCA)
    keep_zero = counts_by_node.sum(0) > 0
    # drop zero branch_lengths (no point to keep it)
    node_branch_zero = branch_lengths.sum(0) > 0
    # combine filters
    keep_node = (keep_zero & node_branch_zero)
    # subset the table (if need, otherwise ignored)
    counts_by_node = counts_by_node[:, keep_node]
    branch_lengths = branch_lengths[keep_node]
    fids = ['n' + i for i in list(tree_index['id'][keep_node].astype(str))]
    tree_index['keep'] = {i: v for i, v in enumerate(keep_node)}
    # re-label tree to return with labels
    tree_relabel = {tid_: tree_index['id_index'][int(tid_[1:])]
                    for tid_ in fids}
    # re-name nodes to match vectorized table
    otu_ids_set = set(otu_ids)
    for new_id, node_ in tree_relabel.items():
        if node_.name in otu_ids_set:
            # replace table name (leaf - nondup)
            fids[fids.index(new_id)] = node_.name
        else:
            # replace tree name (internal)
            node_.name = new_id

    return counts_by_node, tree_index, branch_lengths, fids, otu_ids


def tree_topology_filter(tree, min_depth=DEFAULT_MTD):
    """
    A tree topology filter based on the
    number of descendants. This function
    only removes internal nodes. Tips are
    moved to the parent of the removed node.

    In part, original function comes from
    https://github.com/biocore/wol
    kindly provided here by Qiyun Zhu.
    ----------
    tree : skbio.TreeNode
        tree to calculate metrics
    min_depth: int
        Minimum number of total number of
        descendants (tips) to include a node.
        Default value of zero will retain all nodes.
    Notes
    -----
    The following metrics will be calculated for each node:
    - n : int
        number of descendants (tips)
    Examples
    --------
    >>> # Example from Fig. 9a of Puigbo, et al., 2009, J Biol:
    >>> newick = '((((A,B)n9,C)n8,(D,E)n7)n4,((F,G)n6,(H,I)n5)n3,(J,K)n2)n1;'
    >>> tree = TreeNode.read([newick])
    >>> print(tree.ascii_art())
                                            /-A
                                  /n9------|
                        /n8------|          \\-B
                       |         |
              /n4------|          \\-C
             |         |
             |         |          /-D
             |          \\n7------|
             |                    \\-E
             |
             |                    /-F
    -n1------|          /n6------|
             |         |          \\-G
             |-n3------|
             |         |          /-H
             |          \\n5------|
             |                    \\-I
             |
             |          /-J
              \\n2------|
                        \\-K
    >>> tree_topology_filter(tree, min_depth=2)
    >>> print(tree.ascii_art())
                               /-C
                              |
                     /n8------|--A
                    |         |
                    |          \\-B
           /n4------|
          |         |--D
          |         |
          |          \\-E
          |
          |            /-F
          |            |
    -n1------|         |--G
             |-n3------|
             |         |--H
             |         |
             |         \\-I
             |
             |--J
             |
             \\-K


    """

    # calculate bottom-up metrics
    for node in tree.postorder(include_self=True):
        if node.is_tip():
            node.n = 1
        else:
            children = node.children
            node.n = sum(x.n for x in children)
    # check to ensure tree filters make sense
    # (this has to be done after building the metrics above)
    if tree.root().n <= min_depth:
        raise ValueError('min_depth is equal to tree root value, '
                         'this will result in a table of zero '
                         'features.')
    # non-tip nodes to remove (below the depth filter)
    nodes_to_remove = [node for node in tree.postorder(include_self=True)
                       if (not node.is_tip()) & (node.n <= min_depth)]
    # remove nodes in the tree by moving tips up to parents
    for node_to_remove in nodes_to_remove:
        tips_to_move_up = list(node_to_remove.tips())
        for tip_to_move_up in tips_to_move_up:
            node_to_remove.parent.append(tip_to_move_up)
        node_to_remove.parent.remove(node_to_remove)
    # reconstruct correct topology after removing nodes
    tree.prune()


class build(_BaseConstruct):

    """
    This class can both build N-mode
    tensors from 2D dataframes given
    a count table and mapping data.

    A list of conditional measurements are
    given that identify context measured
    multiple times over the same subjects.
    Additionally a set of subject IDs
    must be provided. Any subjects that are
    missing in a given condition are left
    as completely zero.

    Parameters
    ----------
    table : DataFrame
        table of non-negative count data
        rows = features
        columns = samples
    mapping : DataFrame
        mapping metadata for table
        rows = samples
        columns = categories
    subjects : str, int, or float
        category of sample IDs in metadata
    conditions : str, int, or float
        category of conditional in metadata

    Attributes
    -------
    subject_order : list
        order of subjects in tensor array
    feature_order : list
        order of features in tensor array
    condition_orders : list of lists
        order of conditions for each
        condition in tensor array
    counts : array-like
        Contains table counts.
        N mode tensor of shape
        first dimension = samples
        second dimension = features
        [3..N] dimensions = conditions

    Raises
    ------
    ValueError
        if subject not in mapping
    ValueError
        if any conditions not in mapping
    ValueError
        Table is not 2-dimensional
    ValueError
        Table contains negative values
    ValueError
        Table contains np.inf or -np.inf
    ValueError
        Table contains np.nan or missing.
    Warning
        If a conditional-sample pair
        has multiple IDs associated
        with it the multiple samples
        are meaned.

    References
    ----------
    .. [1] V. Pawlowsky-Glahn, J. J. Egozcue, R. Tolosana-Delgado (2015),
    Modeling and Analysis of Compositional Data, Wiley, Chichester, UK

    .. [2] C. Martino et al., A Novel Sparse Compositional Technique Reveals
    Microbial Perturbations. mSystems. 4 (2019), doi:10.1128/mSystems.00016-19.

    Examples
    --------
    TODO

    """

    def __init__(self):
        """
        Parameters
        ----------
        None

        """

    def construct(self, table, mf, subjects, conditions, branch_lengths=None):
        """
        This function transforms a 2D table
        into a N-Order tensor.

        Parameters
        ----------
        table : DataFrame
            table of non-negative count data
            rows = features
            columns = samples
        mapping : DataFrame
            mapping metadata for table
            rows = samples
            columns = categories
        subjects : str, int, or float
            category of sample IDs in metadata
        conditions : list of strings or ints
            category of conditional in metadata

        Returns
        -------
        self to abstract method

        Raises
        ------
        ValueError
            if subject not in mapping
        ValueError
            if any conditions not in mapping
        ValueError
            Table is not 2-dimensional
        ValueError
            Table contains negative values
        ValueError
            Table contains np.inf or -np.inf
        ValueError
            Table contains np.nan or missing.
        Warning
            If a conditional-sample pair
            has multiple IDs associated
            with it. In this case the
            default method is to mean them.

        Examples
        --------
        TODO

        """

        if subjects not in mf.columns:
            raise ValueError("Subject provided (" +
                             str(subjects) +
                             ") category not in metadata columns.")

        if any(cond_col not in mf.columns for cond_col in conditions):
            missin_cond = ','.join([cond_col for cond_col in conditions
                                    if cond_col not in mf.columns])
            raise ValueError("Conditional category(s) [" +
                             str(missin_cond) +
                             "] not in metadata column(s).")

        if np.count_nonzero(np.isinf(table.values)) != 0:
            raise ValueError('Table contains either np.inf or -np.inf.')

        if np.count_nonzero(np.isnan(table.values)) != 0:
            raise ValueError('Table contains np.nan or missing.')

        if (table.values < 0).any():
            raise ValueError('Table contains negative values.')

        # store all to self
        self.table = table.copy()
        self.feature_order = np.lexsort((table.sum(1).index,
                                         -table.sum(1).values))
        self.feature_order = table.index[self.feature_order]
        self.table = self.table.loc[self.feature_order, :]
        self.mf = mf.copy()
        self.subjects = subjects
        self.conditions = conditions
        self.branch_lengths = branch_lengths
        self._construct()

        return self

    def _construct(self):
        """
        This function forms a tensor
        with missing subject x condition
        pairs left as all zeros.

        Raises
        ------
        Warning
            If a conditional-subject pair
            has multiple samples associated
            with it. In this case the
            default method is to mean them.

        """

        table, mf = self.table, self.mf
        # Step 1: mean samples with multiple conditional overlaps
        col_tmp = [self.subjects] + self.conditions
        duplicated = {k: list(df.index)
                      for k, df in mf.groupby(col_tmp)
                      if df.shape[0] > 1}  # get duplicated conditionals
        if len(duplicated.keys()) > 0:
            duplicated_ids = ','.join(list(set([str(k[0])
                                                for k in duplicated.keys()])))
            warnings.warn(''.join(["Subject(s) (", str(duplicated_ids),
                                   ") contains multiple ",
                                   "samples. Multiple subject counts will be",
                                   " meaned across samples by subject."]),
                          RuntimeWarning)
        for id_, dup in duplicated.items():
            # mean and keep one
            table[dup[0]] = table.loc[:, dup].mean(axis=1).astype(int)
            # drop the other
            table.drop(dup[1:], axis=1)
            mf.drop(dup[1:], axis=0)
        # save direct data
        table_counts = table.values
        rlcr_table = matrix_rclr(table_counts.T,
                                 branch_lengths=self.branch_lengths).T

        # Step 2: fill the tensor (missing are all zero)

        # sort by data within groups to avoid
        # different results with permuted labels
        mf['seq_depth'] = table.sum(0).reindex(mf.index).fillna(0)

        def sortset(grp_col, mf_tmp):
            order_tmp = mf_tmp.groupby(grp_col).sum()
            order_tmp = order_tmp.sort_values('seq_depth').index
            return list(order_tmp)

        # get the ordered subjects
        subject_order = sortset(self.subjects, mf)
        # get un-ordered features (order does not matter)
        feature_order = list(table.index)
        # get the ordered for each conditional
        conditional_orders = [sortset(cond, mf)
                              for cond in self.conditions]
        # generate the dims.
        all_dim = [subject_order]\
            + conditional_orders  # concat all

        # get tensor to fill with counts (all zeros)
        shape = tuple([len(cl) for cl in [subject_order,
                                          feature_order]
                       + all_dim[1:]])
        tensor_counts = np.zeros(tuple(shape))
        tensor_rclr = tensor_counts.copy()

        # generate map from ordered subject and conditions
        # to the original orders in the table
        projection = {
            tuple(
                dim.index(k_) for k_, dim in zip(
                    k, all_dim)): list(
                table.columns).index(
                    df.index[0]) for k, df in mf.groupby(
                        [
                            self.subjects] + self.conditions)}

        # fill the tensor with data
        for T_ind, M_ind in projection.items():
            # get the index from the tensor
            ind_ = [T_ind[:1], list(range(len(table.index)))] + list(T_ind[1:])
            # fill count tensor from table
            tensor_counts[tuple(ind_)] = table_counts[:, M_ind]
            tensor_rclr[tuple(ind_)] = rlcr_table[:, M_ind]
        tensor_rclr[~np.isfinite(tensor_rclr)] = 0.0

        # save metadat and save subject-conditional index
        condition_metadata_map = [{(sid, con): i
                                  for i, sid, con in zip(mf.index,
                                                         mf[self.subjects],
                                                         mf[con])}
                                  for con in self.conditions]
        self.condition_metadata_map = condition_metadata_map
        # save tensor label order
        self.counts = tensor_counts
        self.rclr_transformed = tensor_rclr
        self.subject_order = subject_order
        self.feature_order = feature_order
        self.condition_orders = conditional_orders
        self.mf = self.mf



import os
import click
import pandas as pd
from .__init__ import cli
from biom import load_table
from biom.util import biom_open
from gemelli.preprocessing import TaxonomyError
from gemelli.rpca import rpca as _rpca
from gemelli.rpca import auto_rpca as _auto_rpca
from gemelli.rpca import phylogenetic_rpca as _phylo_rpca
from gemelli._defaults import (DEFAULT_COMP, DEFAULT_MSC, DEFAULT_MTD,
                               DEFAULT_MFC, DEFAULT_OPTSPACE_ITERATIONS,
                               DESC_COMP, DESC_MSC, DESC_MFC,
                               DESC_ITERATIONS, DESC_MINDEPTH,
                               DEFAULT_MFF, DESC_MFF,
                               DESC_COUNTS, DESC_TREE, DESC_TAX_SA)


@cli.command(name='phylogenetic-rpca')
@click.option('--in-biom',
              help=DESC_COUNTS,
              required=True)
@click.option('--in-phylogeny',
              help=DESC_TREE,
              required=True)
@click.option('--taxonomy',
              default=None,
              show_default=True,
              help=DESC_TAX_SA)
@click.option('--output-dir',
              help='Location of output files.',
              required=True)
@click.option('--n-components',
              default=DEFAULT_COMP,
              show_default=True,
              help=DESC_COMP)
@click.option('--min-sample-count',
              default=DEFAULT_MSC,
              show_default=True,
              help=DESC_MSC)
@click.option('--min-feature-count',
              default=DEFAULT_MFC,
              show_default=True,
              help=DESC_MFC)
@click.option('--min-feature-frequency',
              default=DEFAULT_MFF,
              show_default=True,
              help=DESC_MFF)
@click.option('--min-depth',
              default=DEFAULT_MTD,
              show_default=True,
              help=DESC_MINDEPTH)
@click.option('--max-iterations',
              default=DEFAULT_OPTSPACE_ITERATIONS,
              show_default=True,
              help=DESC_ITERATIONS)
def standalone_phylogenetic_rpca(in_biom: str,
                                 in_phylogeny: str,
                                 taxonomy: None,
                                 output_dir: str,
                                 n_components: int,
                                 min_sample_count: int,
                                 min_feature_count: int,
                                 min_feature_frequency: float,
                                 min_depth: int,
                                 max_iterations: int) -> None:
    """Runs phylogenetically informed RPCA with an rclr preprocessing step."""

    # import table
    table = load_table(in_biom)
    # import taxonomy
    taxonomy_table = None
    if taxonomy is not None:
        taxonomy_table = pd.read_csv(taxonomy, sep='\t')
        try:
            taxonomy_table.set_index('Feature ID', inplace=True)
        except KeyError:
            raise TaxonomyError(
                        "Taxonomy file must have a column labled 'Feature ID'."
                        )
    # run the RPCA wrapper
    phylo_res_ = _phylo_rpca(table,
                             in_phylogeny,
                             taxonomy_table,
                             n_components,
                             min_sample_count,
                             min_feature_count,
                             min_feature_frequency,
                             min_depth,
                             max_iterations)
    ord_res, dist_res, phylogeny, counts_by_node, result_taxonomy = phylo_res_
    # If it doesn't already exist, create the output directory.
    # Note that there is technically a race condition here: it's ostensibly
    # possible that some process could delete the output directory after we
    # check that it exists here but before we write the output files to it.
    # However, in this case, we'd just get an error from skbio.io.util.open()
    # (which is called by skbio.OrdinationResults.write()), which makes sense.
    os.makedirs(output_dir, exist_ok=True)

    # write files to output directory
    # Note that this will overwrite files in the output directory that share
    # these filenames (analogous to QIIME 2's behavior if you specify the
    # --o-biplot and --o-distance-matrix options, but differing from QIIME 2's
    # behavior if you specify --output-dir instead).
    ord_res.write(os.path.join(output_dir, 'ordination.txt'))
    dist_res.write(os.path.join(output_dir, 'distance-matrix.tsv'))
    phylogeny.write(os.path.join(output_dir, 'labeled-phylogeny.nwk'))
    if result_taxonomy is not None:
        result_taxonomy.to_csv(os.path.join(output_dir, 't2t-taxonomy.tsv'),
                           sep='\t')
    # write the vectorized count table for Qurro / log-ratios
    with biom_open(os.path.join(output_dir, 'phylo-table.biom'), 'w') as f:
        counts_by_node.to_hdf5(f, "phylo-rpca-count-table")


@cli.command(name='rpca')
@click.option('--in-biom',
              help=DESC_COUNTS,
              required=True)
@click.option('--output-dir',
              help='Location of output files.',
              required=True)
@click.option('--n-components',
              default=DEFAULT_COMP,
              show_default=True,
              help=DESC_COMP)
@click.option('--min-sample-count',
              default=DEFAULT_MSC,
              show_default=True,
              help=DESC_MSC)
@click.option('--min-feature-count',
              default=DEFAULT_MFC,
              show_default=True,
              help=DESC_MFC)
@click.option('--min-feature-frequency',
              default=DEFAULT_MFF,
              show_default=True,
              help=DESC_MFF)
@click.option('--max-iterations',
              default=DEFAULT_OPTSPACE_ITERATIONS,
              show_default=True,
              help=DESC_ITERATIONS)
def standalone_rpca(in_biom: str,
                    output_dir: str,
                    n_components: int,
                    min_sample_count: int,
                    min_feature_count: int,
                    min_feature_frequency: float,
                    max_iterations: int) -> None:
    """Runs RPCA with an rclr preprocessing step."""

    # import table
    table = load_table(in_biom)
    # run the RPCA wrapper
    ord_res, dist_res = _rpca(table,
                              n_components,
                              min_sample_count,
                              min_feature_count,
                              min_feature_frequency,
                              max_iterations)

    # If it doesn't already exist, create the output directory.
    # Note that there is technically a race condition here: it's ostensibly
    # possible that some process could delete the output directory after we
    # check that it exists here but before we write the output files to it.
    # However, in this case, we'd just get an error from skbio.io.util.open()
    # (which is called by skbio.OrdinationResults.write()), which makes sense.
    os.makedirs(output_dir, exist_ok=True)

    # write files to output directory
    # Note that this will overwrite files in the output directory that share
    # these filenames (analogous to QIIME 2's behavior if you specify the
    # --o-biplot and --o-distance-matrix options, but differing from QIIME 2's
    # behavior if you specify --output-dir instead).
    ord_res.write(os.path.join(output_dir, 'ordination.txt'))
    dist_res.write(os.path.join(output_dir, 'distance-matrix.tsv'))


@cli.command(name='auto-rpca')
@click.option('--in-biom',
              help=DESC_COUNTS,
              required=True)
@click.option('--output-dir',
              help='Location of output files.',
              required=True)
@click.option(
    '--min-sample-count',
    default=DEFAULT_MSC,
    show_default=True,
    help=DESC_MSC)
@click.option(
    '--min-feature-count',
    default=DEFAULT_MFC,
    show_default=True,
    help=DESC_MFC)
@click.option(
    '--min-feature-frequency',
    default=DEFAULT_MFF,
    show_default=True,
    help=DESC_MFF)
@click.option(
    '--max-iterations',
    default=DEFAULT_OPTSPACE_ITERATIONS,
    show_default=True,
    help=DESC_ITERATIONS)
def auto_rpca(in_biom: str,
              output_dir: str,
              min_sample_count: int,
              min_feature_count: int,
              min_feature_frequency: float,
              max_iterations: int) -> None:
    """Runs RPCA with an rclr preprocessing step and auto-estimates the
       rank (i.e. n-components parameter)."""

    # import table
    table = load_table(in_biom)
    # run the RPCA wrapper
    ord_res, dist_res = _auto_rpca(table,
                                   min_sample_count,
                                   min_feature_count,
                                   min_feature_frequency,
                                   max_iterations)

    # If it doesn't already exist, create the output directory.
    # Note that there is technically a race condition here: it's ostensibly
    # possible that some process could delete the output directory after we
    # check that it exists here but before we write the output files to it.
    # However, in this case, we'd just get an error from skbio.io.util.open()
    # (which is called by skbio.OrdinationResults.write()), which makes sense.
    os.makedirs(output_dir, exist_ok=True)

    # write files to output directory
    # Note that this will overwrite files in the output directory that share
    # these filenames (analogous to QIIME 2's behavior if you specify the
    # --o-biplot and --o-distance-matrix options, but differing from QIIME 2's
    # behavior if you specify --output-dir instead).
    ord_res.write(os.path.join(output_dir, 'ordination.txt'))
    dist_res.write(os.path.join(output_dir, 'distance-matrix.tsv'))



import pandas as pd
import qiime2
import numpy as np
import pandas.util.testing as pdt
from skbio.util import get_data_path
from qiime2.plugin.testing import TestPluginBase
from gemelli.q2._format import TrajectoryFormat


class TestTrajectoryFormatTransformers(TestPluginBase):
    package = "gemelli.q2.tests"

    def test_pd_series_to_first_differences_format(self):

        transformer = self.get_transformer(pd.DataFrame, TrajectoryFormat)
        test_df = pd.DataFrame(np.random.normal(size=(10, 3)),
                               [str(i) for i in range(10)],
                               ['PC1', 'PC2', 'PC3'])
        test_df.index.name = '#SampleID'
        result = transformer(test_df)
        self.assertIsInstance(result, TrajectoryFormat)

    def test_trajectory_format_to_pd_dataframe(self):
        _, obs = self.transform_format(TrajectoryFormat,
                                       pd.DataFrame,
                                       'trajectory.tsv')
        exp = pd.read_csv(get_data_path('trajectory.tsv'),
                          sep='\t',
                          index_col=0)
        exp.index = exp.index.astype(str)
        exp.columns = exp.columns.astype(str)
        pdt.assert_frame_equal(obs, exp)

    def test_trajectory_format_to_metadata(self):
        _, obs = self.transform_format(TrajectoryFormat,
                                       qiime2.Metadata,
                                       'trajectory.tsv')
        exp = pd.read_csv(get_data_path('trajectory.tsv'),
                          sep='\t',
                          index_col=0)
        exp.index = exp.index.astype(str)
        exp.columns = exp.columns.astype(str)
        exp = qiime2.Metadata(exp)
        self.assertEqual(obs, exp)



# ----------------------------------------------------------------------------
# Copyright (c) 2019--, gemelli development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# ----------------------------------------------------------------------------

# Configuration file where you can set the parameter default values and
# descriptions. This is used by both the standalone RPCA and QIIME 2 RPCA sides
# of gemelli.


DEFAULT_MTD = 0
DEFAULT_BL = None
DEFAULT_COMP = 3
DEFAULT_MSC = 0
DEFAULT_MFC = 0
DEFAULT_MFF = 0
DEFAULT_OPTSPACE_ITERATIONS = 5
DEFAULT_TENSALS_MAXITER = 25
DEFAULT_FMETA = None
DEFAULT_COND = None
DESC_BIN = ("The feature table containing the "
            "samples over which metric should be computed.")
DESC_COUNTS = ("The feature table in biom format containing the "
               "samples over which metric should be computed.")
DESC_TREE = ("Phylogenetic tree containing tip identifiers that "
             "correspond to the feature identifiers in the table. "
             "This tree can contain tip ids that are not present "
             "in the table, but all feature ids in the table must "
             "be present in this tree.")
DESC_MINDEPTH = ("Minimum number of total number of "
                 "descendants (tips) to include a node. "
                 "Default value of zero will retain all nodes "
                 "(including tips).")
DESC_COMP = ("The underlying low-rank structure."
             " The input can be an integer "
             "(suggested: 1 < rank < 10) [minimum 2]."
             " Note: as the rank increases the runtime"
             " will increase dramatically.")
DESC_ITERATIONS = ("The number of iterations to optimize the solution"
                   " (suggested to be below 100; beware of overfitting)"
                   " [minimum 1]")
DESC_INIT = ("The number of initialization vectors. Larger values will"
             "give more accurate factorization but will be more "
             "computationally expensive [minimum 1]")
DESC_ITERATIONSALS = ("Max number of Alternating Least Square (ALS)"
                      " optimization iterations (suggested to be below 100;"
                      " beware of overfitting) [minimum 1]")
DESC_ITERATIONSRTPM = ("Max number of Robust Tensor Power Method (RTPM)"
                       " optimization iterations (suggested to be below 100;"
                       " beware of overfitting) [minimum 1]")
DESC_MSC = ("Minimum sum cutoff of sample across all features. "
            "The value can be at minimum zero and must be an whole"
            " integer. It is suggested to be greater than or equal"
            " to 500.")
DESC_MFC = ("Minimum sum cutoff of features across all samples. "
            "The value can be at minimum zero and must be an whole"
            " integer")
DESC_MFF = ("Minimum percentage of samples a feature must appear"
            " with a value greater than zero. This value can range"
            " from 0 to 100 with decimal values allowed.")
DESC_OUT = "Location of output files."
DESC_FMETA = "Feature metadata file in QIIME2 formatting."
DESC_SMETA = "Sample metadata file in QIIME2 formatting."
DESC_TAX_Q2 = ("Taxonomy file in QIIME2 formatting.")
DESC_TAX_SA = ("A tsv taxonomy file which contains a column labeled either "
               "'taxon' or 'taxonomy' (case insensitive) and the first column "
               "must be labeled "
               "'Feature ID'. A new taxonomy file will be created.")
DESC_T2T_TAX = ("The resulting tax2Tree taxonomy and will include taxonomy "
                "for both internal nodes and tips.")
DESC_STBL = ("A table with samples aggergated by subject. "
             "This can be used and input into Empress.")
DESC_SUBJ = ("Metadata column containing subject IDs to"
             " use for pairing samples. WARNING: if"
             " replicates exist for an individual ID at"
             " either state_1 to state_N, that subject will"
             " be mean grouped by default.")
DESC_COND = ("Metadata column containing state (e.g.,Time, BodySite)"
             " across which samples are paired."
             " At least one is required but up to four are allowed"
             " by other state inputs.")
QORD = ("A trajectory is an ordination that can be visualized "
        "over time or another context.")
QDIST = ("A sample-sample distance matrix generated from "
         " the euclidean distance of the subject-state "
         "ordinations and itself.")
QLOAD = ("Compositional biplot of subjects as points and features as arrows."
         " Where the variation between subject groupings is explained by the"
         " log-ratio between opposing arrows. "
         "WARNING: The % variance explained is only spread over n_components "
         "and can be inflated.")
QSOAD = ("Compositional biplot of states as points and features as arrows."
         " Where the variation between subject groupings is explained by the"
         " log-ratio between opposing arrows. "
         "WARNING: The % variance explained is only spread over n_components "
         "and can be inflated.")
QTREE = ("The input tree with all nodes matched in name to the features "
         "in the counts_by_node table.")
QTREECOUNT = ("A table with all tree internal nodes as features with the "
              "sum of all children of that node (i.e. Fast UniFrac).")
QBIPLOT = "A biplot of the (Robust Aitchison) RPCA feature loadings"
QADIST = "The Aitchison distance of the sample loadings from RPCA."
QRCLR = "A rclr transformed table. Note: zero/missing values have NaNs"



import unittest
import numpy as np
from scipy.linalg import qr
from gemelli.factorization import tenals


class Testtenals(unittest.TestCase):

    def setUp(self):
        # generate random noiseless low-rank orthogonal tensor
        r = 3  # rank is 2
        n1 = 10
        n2 = 12
        n3 = 8
        U01 = np.random.rand(n1, r)
        U02 = np.random.rand(n2, r)
        U03 = np.random.rand(n3, r)
        U1, temp = qr(U01)
        U2, temp = qr(U02)
        U3, temp = qr(U03)
        U1 = U1[:, 0:r]
        U2 = U2[:, 0:r]
        U3 = U3[:, 0:r]
        T = np.zeros((n1, n2, n3))
        for i in range(n3):
            T[:, :, i] = np.matmul(U1, np.matmul(np.diag(U3[i, :]), U2.T))
        # sample entries
        p = 2 * (r ** 0.5 * np.log(n1 * n2 * n3)) / np.sqrt(n1 * n2 * n3)
        self.E = abs(np.ceil(np.random.rand(n1, n2, n3) - 1 + p))
        self.TE = T * self.E
        self.TE_noise = self.TE + (0.0001 / np.sqrt(n1 * n2 * n3)
                                   * np.random.randn(n1, n2, n3) * self.E)
        self.n3 = n3
        self.U1 = U1
        self.U2 = U2
        self.U3 = U3

    def test_tenals_noiseless(self):
        # TenAls no noise
        loadings, s, dist = tenals(self.TE, self.E)
        L1, L2, L3 = loadings
        s = np.diag(s)
        # test accuracy
        rmse = 0
        for i3 in range(self.n3):
            A1 = self.U1
            A2 = np.matmul(self.U2, np.diag(self.U3[i3, :]))
            B1 = L1
            B2 = np.matmul(L2, np.diag(L3[i3, :] * s.T.flatten()))
            rmse += np.trace(np.matmul(np.matmul(A1.T, A1), np.matmul(A2.T,
                                                                      A2))) + \
                np.trace(np.matmul(np.matmul(B1.T, B1), np.matmul(B2.T,
                                                                  B2))) + \
                -2 * np.trace(np.matmul(np.matmul(B1.T, A1), np.matmul(A2.T,
                                                                       B2)))
        self.assertTrue(1e2 > abs(rmse))

    def test_tenals_noise(self):
        # TenAls no noise
        loadings, s, dist = tenals(self.TE_noise, self.E)
        L1, L2, L3 = loadings
        s = np.diag(s)
        # test accuracy
        rmse = 0
        for i3 in range(self.n3):
            A1 = self.U1
            A2 = np.matmul(self.U2, np.diag(self.U3[i3, :]))
            B1 = L1
            B2 = np.matmul(L2, np.diag(L3[i3, :] * s.T.flatten()))
            rmse += np.trace(np.matmul(np.matmul(A1.T, A1), np.matmul(A2.T,
                                                                      A2))) + \
                np.trace(np.matmul(np.matmul(B1.T, B1), np.matmul(B2.T,
                                                                  B2))) + \
                -2 * np.trace(np.matmul(np.matmul(B1.T, A1), np.matmul(A2.T,
                                                                       B2)))
        self.assertTrue(1e2 > abs(rmse))



import unittest
from gemelli.base import _BaseConstruct, _BaseImpute


class Test_BaseConstruct(unittest.TestCase):
    def test_no_instantiation(self):
        class Foo(_BaseConstruct):
            pass


class Test_BaseImpute(unittest.TestCase):
    def test_no_instantiation(self):
        class Foo_boo(_BaseImpute):
            pass



from __future__ import division
# utils
import pandas as pd
import numpy as np
# blocks
from scipy.stats import norm
from numpy.random import poisson, lognormal
from skbio.stats.composition import closure
# Set random state
rand = np.random.RandomState(42)


def Homoscedastic(X_noise, intensity):
    """ uniform normally dist. noise """
    X_noise = np.array(X_noise)
    err = intensity * np.ones_like(X_noise.copy())
    X_noise = rand.normal(X_noise.copy(), err)

    return X_noise


def Heteroscedastic(X_noise, intensity):
    """ non-uniform normally dist. noise """
    err = intensity * np.ones_like(X_noise)
    i = rand.randint(0, err.shape[0], 5000)
    j = rand.randint(0, err.shape[1], 5000)
    err[i, j] = intensity
    X_noise = abs(rand.normal(X_noise, err))

    return X_noise


def Subsample(X_noise, spar, num_samples):
    """ yij ~ PLN( lambda_{ij}, /phi ) """
    # subsample
    mu = spar * closure(X_noise.T).T
    X_noise = np.vstack([poisson(lognormal(np.log(mu[:, i]), 1))
                         for i in range(num_samples)]).T
    # add sparsity

    return X_noise


def block_diagonal_gaus(
        ncols,
        nrows,
        nblocks,
        overlap=0,
        minval=0,
        maxval=1.0):
    """
    Generate block diagonal with Gaussian distributed values within blocks.

    Parameters
    ----------

    ncol : int
        Number of columns

    nrows : int
        Number of rows

    nblocks : int
        Number of blocks, mucst be greater than one

    overlap : int
        The Number of overlapping columns (Default = 0)

    minval : int
        The min value output of the table (Default = 0)

    maxval : int
        The max value output of the table (Default = 1)


    Returns
    -------
    np.array
        Table with a block diagonal where the rows represent samples
        and the columns represent features.  The values within the blocks
        are gaussian distributed between 0 and 1.
    Note
    ----
    The number of blocks specified by `nblocks` needs to be greater than 1.

    """

    if nblocks <= 1:
        raise ValueError('`nblocks` needs to be greater than 1.')
    mat = np.zeros((nrows, ncols))
    gradient = np.linspace(0, 10, nrows)
    mu = np.linspace(0, 10, ncols)
    sigma = 1
    xs = [norm.pdf(gradient, loc=mu[i], scale=sigma)
          for i in range(len(mu))]
    mat = np.vstack(xs).T

    block_cols = ncols // nblocks
    block_rows = nrows // nblocks
    for b in range(nblocks - 1):

        gradient = np.linspace(
            5, 5, block_rows)  # samples (bock_rows)
        # features (block_cols+overlap)
        mu = np.linspace(0, 10, block_cols + overlap)
        sigma = 2.0
        xs = [norm.pdf(gradient, loc=mu[i], scale=sigma)
              for i in range(len(mu))]

        B = np.vstack(xs).T * maxval
        lower_row = block_rows * b
        upper_row = min(block_rows * (b + 1), nrows)
        lower_col = block_cols * b
        upper_col = min(block_cols * (b + 1), ncols)

        if b == 0:
            mat[lower_row:upper_row,
                lower_col:int(upper_col + overlap)] = B
        else:
            ov_tmp = int(overlap / 2)
            if (B.shape) == (mat[lower_row:upper_row,
                                 int(lower_col - ov_tmp):
                                 int(upper_col + ov_tmp + 1)].shape):
                mat[lower_row:upper_row, int(
                    lower_col - ov_tmp):int(upper_col + ov_tmp + 1)] = B
            elif (B.shape) == (mat[lower_row:upper_row,
                                   int(lower_col - ov_tmp):
                                   int(upper_col + ov_tmp)].shape):
                mat[lower_row:upper_row, int(
                    lower_col - ov_tmp):int(upper_col + ov_tmp)] = B
            elif (B.shape) == (mat[lower_row:upper_row,
                                   int(lower_col - ov_tmp):
                                   int(upper_col + ov_tmp - 1)].shape):
                mat[lower_row:upper_row, int(
                    lower_col - ov_tmp):int(upper_col + ov_tmp - 1)] = B

    upper_col = int(upper_col - overlap)
    # Make last block fill in the remainder
    gradient = np.linspace(5, 5, nrows - upper_row)
    mu = np.linspace(0, 10, ncols - upper_col)
    sigma = 4
    xs = [norm.pdf(gradient, loc=mu[i], scale=sigma)
          for i in range(len(mu))]
    B = np.vstack(xs).T * maxval

    mat[upper_row:, upper_col:] = B

    return mat


def build_block_model(
        rank,
        hoced,
        hsced,
        spar,
        C_,
        num_samples,
        num_features,
        overlap=0,
        mapping_on=True):
    """
    Generates hetero and homo scedastic noise on base truth block
    diagonal with Gaussian distributed values within blocks.

    Parameters
    ----------

    rank : int
        Number of blocks


    hoced : int
        Amount of homoscedastic noise

    hsced : int
        Amount of heteroscedastic noise

    inten : int
        Intensity of the noise

    spar : int
        Level of sparsity

    C_ : int
        Intensity of real values

    num_features : int
        Number of rows

    num_samples : int
        Number of columns

    overlap : int
        The Number of overlapping columns (Default = 0)

    mapping_on : bool
        if true will return pandas dataframe mock mapping file by block


    Returns
    -------
    Pandas Dataframes
    Table with a block diagonal where the rows represent samples
    and the columns represent features.  The values within the blocks
    are gaussian.

    Note
    ----
    The number of blocks specified by `nblocks` needs to be greater than 1.

    """

    # make a mock OTU table
    X_true = block_diagonal_gaus(
        num_samples,
        num_features,
        rank,
        overlap,
        minval=.01,
        maxval=C_)
    if mapping_on:
        # make a mock mapping data
        mappning_ = pd.DataFrame(np.array([['Cluster %s' % str(x)] *
                                           int(num_samples / rank)
                                           for x in range(1,
                                           rank + 1)]).flatten(),
                                 columns=['example'],
                                 index=['sample_' + str(x)
                                        for x in range(0, num_samples - 2)])

    X_noise = X_true.copy()
    X_noise = np.array(X_noise)
    # add Homoscedastic noise
    X_noise = Homoscedastic(X_noise, hoced)
    # add Heteroscedastic noise
    X_noise = Heteroscedastic(X_noise, hsced)
    # Induce low-density into the matrix
    X_noise = Subsample(X_noise, spar, num_samples)

    # return the base truth and noisy data
    if mapping_on:
        return X_true, X_noise, mappning_
    else:
        return X_true, X_noise



import unittest
import numpy as np
from gemelli.matrix_completion import MatrixCompletion
from gemelli.preprocessing import matrix_rclr
from gemelli.optspace import OptSpace
from skbio.stats.composition import clr
from gemelli.q2.tests.test_rpca_method import create_test_table
from nose.tools import nottest


@nottest
def rclr_test_table():
    # build a table to test
    test_table = create_test_table()
    # export table from biom
    test_table = test_table.matrix_data.toarray()
    # the matrix_rclr is tested in other places
    # this is just used as input into
    # the OptSpace tests
    test_table = np.array(test_table)
    table_rclr = matrix_rclr(test_table)

    return test_table, table_rclr


class TestOptSpace(unittest.TestCase):

    def setUp(self):
        self.test_table, self.test_rclr = rclr_test_table()
        self.rank = 2
        self.iteration = 5
        self.tol = 1e-5

    def test_OptSpace(self):
        """Tests the basic validity of the
        actual OptSpace() method's output."""

        # run base OptSpace
        opt = MatrixCompletion(n_components=self.rank,
                               max_iterations=self.iteration,
                               tol=self.tol).fit(self.test_rclr)
        U_res = MatrixCompletion(n_components=self.rank,
                                 max_iterations=self.iteration,
                                 tol=self.tol).fit_transform(self.test_rclr)
        # use base optspace helper to check
        # that wrapper is not changing outcomes
        U_exp, s_exp, V_exp = OptSpace(n_components=self.rank,
                                       max_iterations=self.iteration,
                                       tol=self.tol).solve(self.test_rclr)
        # more exact testing of directionally is done
        # in test_method.py. Here we just compare abs
        # see  (c/o @cameronmartino's comment in #29).
        for i in range(self.rank):
            np.testing.assert_array_almost_equal(abs(U_exp[:, i]),
                                                 abs(opt.sample_weights[:, i]))
            np.testing.assert_array_almost_equal(abs(s_exp[:, i]),
                                                 abs(opt.s[:, i]))
            np.testing.assert_array_almost_equal(
                abs(V_exp[:, i]), abs(opt.feature_weights[:, i]))
            np.testing.assert_array_almost_equal(abs(U_exp[:, i]),
                                                 abs(U_res[:, i]))

    def test_OptSpace_rank_low_raises(self):
        """Tests ValueError for OptSpace() rank too low."""
        try:
            MatrixCompletion(n_components=1).fit(self.test_rclr)
        except ValueError:
            pass
        else:
            raise AssertionError("ValueError was not raised")

    def test_OptSpace_rank_high_raises(self):
        """Tests ValueError for OptSpace() rank too high."""
        try:
            MatrixCompletion(n_components=10000).fit(self.test_rclr)
        except ValueError:
            pass
        else:
            raise AssertionError("ValueError was not raised")

    def test_OptSpace_rank_mid_raises(self):
        """Tests ValueError for OptSpace() rank too mid."""
        try:
            MatrixCompletion(n_components=100).fit(self.test_rclr)
        except ValueError:
            pass
        else:
            raise AssertionError("ValueError was not raised")

    def test_OptSpace_iter_raises(self):
        """Tests ValueError for OptSpace() iteration 0."""
        # test iter too low
        try:
            MatrixCompletion(max_iterations=0).fit(self.test_rclr)
        except ValueError:
            pass
        else:
            raise AssertionError("ValueError was not raised")

    def test_OptSpace_illformatted_raises(self):
        """Tests ValueError for OptSpace() no infs."""
        # test inf
        try:
            MatrixCompletion().fit(clr(self.test_table))
        except ValueError:
            pass
        else:
            raise AssertionError("ValueError was not raised")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

# ----------------------------------------------------------------------------
# Copyright (c) 2019--, gemelli development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# ----------------------------------------------------------------------------

from setuptools.command.egg_info import egg_info
from setuptools.command.develop import develop
from setuptools.command.install import install
import re
import ast
import os
from setuptools import find_packages, setup

# Dealing with Cython
USE_CYTHON = os.environ.get('USE_CYTHON', False)
ext = '.pyx' if USE_CYTHON else '.c'

# bootstrap numpy intall
# https://stackoverflow.com/questions/51546255/
# python-package-setup-setup-py-with-customisation
# -to-handle-wrapped-fortran


def custom_command():
    import sys
    if sys.platform in ['darwin', 'linux']:
        os.system('pip install numpy')


class CustomInstallCommand(install):
    def run(self):
        install.run(self)
        custom_command()


class CustomDevelopCommand(develop):
    def run(self):
        develop.run(self)
        custom_command()


class CustomEggInfoCommand(egg_info):
    def run(self):
        egg_info.run(self)
        custom_command()


extensions = [
]

if USE_CYTHON:
    from Cython.Build import cythonize
    extensions = cythonize(extensions)

classes = """
    Development Status :: 3 - Alpha
    License :: OSI Approved :: BSD License
    Topic :: Software Development :: Libraries
    Topic :: Scientific/Engineering
    Topic :: Scientific/Engineering :: Bio-Informatics
    Programming Language :: Python :: 3
    Programming Language :: Python :: 3 :: Only
    Operating System :: Unix
    Operating System :: POSIX
    Operating System :: MacOS :: MacOS X
"""
classifiers = [s.strip() for s in classes.split('\n') if s]

description = ('Robust Aitchison Tensor Decomposition for sparse count data')

with open('README.md') as f:
    long_description = f.read()

# version parsing from __init__ pulled from Flask's setup.py
# https://github.com/mitsuhiko/flask/blob/master/setup.py
_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('gemelli/__init__.py', 'rb') as f:
    hit = _version_re.search(f.read().decode('utf-8')).group(1)
    version = str(ast.literal_eval(hit))

standalone = ['gemelli=gemelli.scripts.__init__:cli']
q2cmds = ['q2-gemelli=gemelli.q2.plugin_setup:plugin']

setup(name='gemelli',
      version=version,
      license='BSD-3-Clause',
      description=description,
      long_description=long_description,
      long_description_content_type='text/markdown',
      author="gemelli development team",
      author_email="cmartino@eng.ucsd.edu",
      maintainer="gemelli development team",
      maintainer_email="cmartino@eng.ucsd.edu",
      packages=find_packages(),
      ext_modules=extensions,
      install_requires=[
          'numpy >= 1.12.1',
          'click',
          'pandas >= 0.10.0',
          'scipy >= 0.19.1',
          'nose >= 1.3.7',
          'scikit-learn >= 0.18.1',
          'scikit-bio > 0.5.3',
          'biom-format',
          'h5py',
          'iow',
          'tax2tree'],
      classifiers=classifiers,
      entry_points={'qiime2.plugins': q2cmds,
                    'console_scripts': standalone},
      # Inclusion of citations.bib in package_data based on how this is done in
      # q2-emperor's setup.py file
      package_data={'gemelli': ['citations.bib']},
      cmdclass={'install': CustomInstallCommand,
                'develop': CustomDevelopCommand,
                'egg_info': CustomEggInfoCommand, },
      zip_safe=False)



# ----------------------------------------------------------------------------
# Copyright (c) 2019--, gemelli development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# ----------------------------------------------------------------------------

import unittest
from pandas.testing import assert_series_equal


def assert_ordinationresults_equal(o1, o2, precision=5, verbose=False):
    """Asserts rough equality of OrdinationResults.

    Parameters
    ----------
        o1: skbio.OrdinationResults
        o2: skbio.OrdinationResults
            The values of these two objects will be compared.
            The following guidelines are followed when comparing o1.features
            and o2.features, and when comparing o1.samples and o2.samples:

                -The index names must be equal, but they can be in any order
                 (they'll be sorted before we compare the DataFrames). However,
                 it is NOT OK for these indices to contain repeated index
                 names; if this is the case, this function will fail.

                -The number of columns must be the same, but their names don't
                 matter.

                -For each column (aka pandas Series) within the DataFrames, we
                 run the pandas.testing.assert_series_equal() function (so
                 o1.features' first column is compared with o2.features' first
                 column, and so on). This function is called with
                 check_names=False (since we allow column names to be
                 different) and check_less_precise=[precision] (where
                 [precision] is the optional argument to this function
                 described below). We also wrap this invocation within a
                 try/except block, to allow for the following:

                -If the first invocation of assert_series_equal() fails between
                 two columns, we try calling it again with one of the columns
                 negated. This is because negating an entire axis in an
                 ordination doesn't really change the interpretation; see
                 @cameronmartino's comments on PR #29 in the biocore/DEICODE
                 repo for a better explanation of this.

        precision: int
            The argument to be passed to pandas.testing.assert_series_equal()
            when comparing the columns within o1 and o2's feature and sample
            DataFrames.

        verbose: bool
            If True, will print some information as this goes through the
            columns.

    """
    # This is just a tuple where the DataFrames within each 2-tuple within
    # it will be compared
    res_exp = ((o1.features, o2.features, "feature"),
               (o1.samples, o2.samples, "sample"))

    for (res, exp, aspect) in res_exp:
        # Just to protect ourselves, ensure that the indices only contain
        # unique names.
        assert len(set(res.index)) == len(res.index)
        assert len(set(exp.index)) == len(exp.index)
        # Row order doesn't matter, but the row names should match
        assert set(res.index) == set(exp.index)
        # Column names don't matter, but order does
        assert len(res.columns) == len(exp.columns)
        # Now, we can actually check that the ordination values match.
        # (First, we sort so that the rows in both the result and expected
        # DFs are in the same order, to enable us to use check_names=False
        # when calling assert_series_equal().)
        res_sorted = res.sort_index()
        exp_sorted = exp.sort_index()
        for col_index in range(len(res_sorted.columns)):
            # Extract the n-th column (a pandas Series) from both the
            # result and expected DataFrame, then compare their values.
            res_series = res_sorted.iloc[:, col_index]
            exp_series = exp_sorted.iloc[:, col_index]
            try:
                # First, just try comparing the two PCs and seeing if their
                # values are approximately equal.
                assert_series_equal(res_series, exp_series, check_names=False,
                                    check_less_precise=precision)
            except AssertionError:
                # It's fine for any of the "PC"s (i.e. columns in the
                # OrdinationResults) to be off by a factor of -1, since
                # that doesn't really change the interpretation of anything
                # (c/o @cameronmartino's comment in #29).
                # To allow for this case to pass the tests, we just try
                # negating one of the series, and seeing if
                # that makes them approximately equal.
                # (If they're *still* not equal, this test will fail.)
                assert_series_equal(-res_series, exp_series, check_names=False,
                                    check_less_precise=precision)
            if verbose:
                print("PC {} for the {} ordination matches.".format(col_index,
                                                                    aspect))


class CliTestCase(unittest.TestCase):

    def assertExitCode(self, value, result):
        try:
            self.assertEqual(0, result.exit_code)
        except AssertionError:
            ex = result.exception
            error = Exception('Command failed with non-zero exit code')
            raise error.with_traceback(ex.__traceback__)


def absolute_sort(matt):
    matt = abs(matt)
    matt = matt[matt[:, 0].argsort()]
    return matt



import qiime2.plugin.model as model
from qiime2.plugin import ValidationError


class TrajectoryFormat(model.TextFileFormat):
    def _validate(self, n_records=None):
        with self.open() as fh:
            # check the header column names
            header = fh.readline()
            comp_columns = [i for i, head in enumerate(header.split('\t'))
                            if 'PC' in head]
            # ensure there at least two components
            if len(comp_columns) < 1:
                raise ValidationError('No PC# columns present. '
                                      'There should be at least one PC#'
                                      '(i.e. at minimum PC1) in trajectory.')
            # validate the body of the data
            for line_number, line in enumerate(fh, start=2):
                cells = line.split('\t')
                pc_type = [is_float(cells[c].strip()) for c in comp_columns]
                if not all(pc_type):
                    raise ValidationError('Non float values in trajectory.')
                if n_records is not None and (line_number - 1) >= n_records:
                    break

    def _validate_(self, level):
        record_count_map = {'min': 5, 'max': None}
        self._validate(record_count_map[level])


def is_float(str):
    try:
        float(str)
        return True
    except ValueError:
        return False


TrajectoryDirectoryFormat = model.SingleFileDirectoryFormat(
    'TrajectoryDirectoryFormat', 'trajectory.tsv',
    TrajectoryFormat)



# ----------------------------------------------------------------------------
# Copyright (c) 2019--, gemelli development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# ----------------------------------------------------------------------------

from abc import abstractmethod


class _BaseImpute(object):

    """Base class for imputation methods.
    Warning: This class should not be used directly.
    Use derived classes instead.
    """
    @abstractmethod
    def fit(self):
        """ Placeholder for fit this
        should be implemetned by sub-method"""
    @abstractmethod
    def label(self):
        """ Placeholder for fit this
        should be implemetned by sub-method"""


class _BaseConstruct(object):

    """Base class for transformation/norm methods.
    Warning: This class should not be used directly.
    Use derived classes instead.
    """
    @abstractmethod
    def construct(self):
        """
        conditional_loading  : array-like or list of array-like
             The conditional loading vectors
             of shape (conditions, r) if there is 1 type
             of condition, and a list of such matrices if
             there are more than 1 type of condition
         feature_loading : array-like
             The feature loading vectors
             of shape (features, r)
         sample_loading : array-like
             The sample loading vectors
             of shape (samples, r) """



# ----------------------------------------------------------------------------
# Copyright (c) 2019--, gemelli development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# ----------------------------------------------------------------------------

import numpy as np
from gemelli.optspace import OptSpace
from .base import _BaseImpute
from scipy.spatial import distance


class MatrixCompletion(_BaseImpute):

    def __init__(self, n_components=2, max_iterations=5, tol=1e-5):
        """

        This form of matrix completion uses OptSpace (1). Furthermore,
        here we directly interpret the loadings generated from matrix
        completion as a dimensionality reduction.

        Parameters
        ----------

        X: numpy.ndarray - a rclr preprocessed matrix of shape (M,N)
        N = Features (i.e. OTUs, metabolites)
        M = Samples

        n_components: int, optional : Default is 2
        The underlying rank of the default set
        to 2 as the default to prevent overfitting.

        max_iterations: int, optional : Default is 5
        The number of convex iterations to optimize the solution
        If iteration is not specified, then the default iteration is 5.
        Which redcues to a satisfactory error threshold.

        tol: float, optional : Default is 1e-5
        Error reduction break, if the error reduced is
        less than this value it will return the solution

        Returns
        -------
        U: numpy.ndarray - "Sample Loadings" or the unitary matrix
        having left singular vectors as columns. Of shape (M,n_components)

        s: numpy.ndarray - The singular values,
        sorted in non-increasing order. Of shape (n_components,n_components).

        V: numpy.ndarray - "Feature Loadings" or Unitary matrix
        having right singular vectors as rows. Of shape (N,n_components)

        solution: numpy.ndarray - (U*S*V.transpose()) of shape (M,N)

        distance: numpy.ndarray - Distance between each
        pair of the two collections of inputs. Of shape (M,M)

        Raises
        ------
        ValueError

        ValueError
            `ValueError: n_components must be at least 2`.

        ValueError
            `ValueError: max_iterations must be at least 1`.

        ValueError
            `ValueError: Data-table contains either np.inf or -np.inf`.

        ValueError
            `ValueError: The n_components must be less
             than the minimum shape of the input table`.

        References
        ----------
        .. [1] Keshavan RH, Oh S, Montanari A. 2009. Matrix completion
                from a few entries (2009_ IEEE International
                Symposium on Information Theory

        Examples
        --------
        TODO

        """

        self.n_components = n_components
        self.max_iterations = max_iterations
        self.tol = tol

        return

    def fit(self, X, y=None):
        """
        Fit the model to X_sparse
        """

        X_sparse = X.copy().astype(np.float64)
        self.X_sparse = X_sparse
        self._fit()
        return self

    def _fit(self):

        # make copy for imputation, check type
        X_sparse = self.X_sparse
        n, m = X_sparse.shape

        # make sure the data is sparse (otherwise why?)
        if np.count_nonzero(np.isinf(X_sparse)) != 0:
            raise ValueError('Contains either np.inf or -np.inf')

        # test n-iter
        if self.max_iterations < 1:
            raise ValueError("max_iterations must be at least 1")

        # check the settings for n_components
        if isinstance(self.n_components, str) and \
           self.n_components.lower() == 'auto':
            # estimate the rank of the matrix
            self.n_components = 'auto'
        # check hardset values
        elif isinstance(self.n_components, int):
            if self.n_components > (min(n, m) - 1):
                raise ValueError("n-components must be at most"
                                 " 1 minus the min. shape of the"
                                 " input matrix.")
            if self.n_components < 2:
                raise ValueError("n-components must "
                                 "be at least 2")
        # otherwise rase an error.
        else:
            raise ValueError("n-components must be "
                             "an interger or 'auto'")

        # return solved matrix
        self.U, self.s, self.V = OptSpace(n_components=self.n_components,
                                          max_iterations=self.max_iterations,
                                          tol=self.tol).solve(X_sparse)
        # save the solution (of the imputation)
        self.solution = self.U.dot(self.s).dot(self.V.T)
        self.eigenvalues = np.diag(self.s)
        self.explained_variance_ratio = list(
            self.eigenvalues / self.eigenvalues.sum())
        self.distance = distance.cdist(self.U, self.U)
        self.feature_weights = self.V
        self.sample_weights = self.U

    def fit_transform(self, X, y=None):
        """
        Returns the final SVD of

        U: numpy.ndarray - "Sample Loadings" or the
        unitary matrix having left singular
        vectors as columns. Of shape (M,n_components)

        s: numpy.ndarray - The singular values,
        sorted in non-increasing order. Of shape (n_components,n_components).

        V: numpy.ndarray - "Feature Loadings" or Unitary matrix
        having right singular vectors as rows. Of shape (N,n_components)

        """
        X_sparse = X.copy().astype(np.float64)
        self.X_sparse = X_sparse
        self._fit()
        return self.sample_weights



import numpy as np
import pandas as pd
from qiime2 import Metadata
from .plugin_setup import plugin
from ._format import TrajectoryFormat


@plugin.register_transformer
def _1(data: pd.DataFrame) -> (TrajectoryFormat):
    ff = TrajectoryFormat()
    with ff.open() as fh:
        data.to_csv(fh, sep='\t', header=True, na_rep=np.nan)
    return ff


@plugin.register_transformer
def _2(ff: TrajectoryFormat) -> (pd.DataFrame):
    # with ff.open() as fh:
    return Metadata.load(str(ff)).to_dataframe()


@plugin.register_transformer
def _3(ff: TrajectoryFormat) -> (Metadata):
    # with ff.open() as fh:
    return Metadata.load(str(ff))



import unittest
import pandas as pd
from skbio import TreeNode
from biom import load_table
from os.path import sep as os_path_sep
from click.testing import CliRunner
from skbio.util import get_data_path
from numpy.testing import (assert_allclose,
                           assert_array_almost_equal)
from gemelli.testing import (CliTestCase,
                             absolute_sort)
from gemelli.scripts._standalone_ctf import (standalone_ctf,
                                             standalone_phylogenetic_ctf)


class Test_standalone_ctf(unittest.TestCase):
    def setUp(self):
        pass

    def test_standalone_ctf(self):
        """Checks the output produced by gemelli's standalone CTF script.

           This is more of an "integration test" than a unit test -- the
           details of the algorithm used by the standalone CTF script are
           checked in more detail in gemelli/tests/test_factorization.py.
        """
        in_table = get_data_path('test-small.biom')
        in_meta = get_data_path('test-small.tsv')
        out_ = os_path_sep.join(in_table.split(os_path_sep)[:-1])
        runner = CliRunner()
        result = runner.invoke(standalone_ctf,
                               ['--in-biom',
                                in_table,
                                '--sample-metadata-file',
                                in_meta,
                                '--individual-id-column',
                                'host_subject_id',
                                '--state-column-1',
                                'context',
                                '--output-dir',
                                out_])
        # check exit code was 0 (indicating success)
        CliTestCase().assertExitCode(0, result)
        # Read the results
        samp_res = pd.read_csv(
            get_data_path('context-subject-ordination.tsv'),
            sep='\t',
            index_col=0)
        feat_res = pd.read_csv(
            get_data_path('context-features-ordination.tsv'),
            sep='\t',
            index_col=0)
        # Read the expected results
        samp_exp = pd.read_csv(
            get_data_path('expected-context-subject-ordination.tsv'),
            sep='\t',
            index_col=0)
        feat_exp = pd.read_csv(
            get_data_path('expected-context-features-ordination.tsv'),
            sep='\t',
            index_col=0)
        # Check that the distance matrix matches our expectations
        comp_col = ['PC1', 'PC2', 'PC3']
        cent_ = samp_res[comp_col].mean().values.max()
        self.assertAlmostEqual(cent_, 0)
        cent_ = feat_res[comp_col].mean().values.max()
        self.assertAlmostEqual(cent_, 0)
        # check matched
        assert_allclose(absolute_sort(samp_res[comp_col].values),
                        absolute_sort(samp_exp[comp_col].values),
                        atol=.5)
        assert_allclose(absolute_sort(feat_res[comp_col].values),
                        absolute_sort(feat_exp[comp_col].values),
                        atol=.5)

    def test_standalone_phylogenetic_ctf(self):
        """Checks the output produced by gemelli's standalone phylogenetic CTF.

           This is more of an "integration test" than a unit test -- the
           details of the algorithm used by the standalone CTF script are
           checked in more detail in gemelli/tests/test_factorization.py.
        """
        in_table = get_data_path('test-small.biom')
        in_meta = get_data_path('test-small.tsv')
        in_tree = get_data_path('tree.nwk')
        out_ = os_path_sep.join(in_table.split(os_path_sep)[:-1])
        runner = CliRunner()
        result = runner.invoke(standalone_phylogenetic_ctf,
                               ['--in-biom',
                                in_table,
                                '--in-phylogeny',
                                in_tree,
                                '--sample-metadata-file',
                                in_meta,
                                '--individual-id-column',
                                'host_subject_id',
                                '--state-column-1',
                                'context',
                                '--output-dir',
                                out_])
        # check exit code was 0 (indicating success)
        CliTestCase().assertExitCode(0, result)
        # Read the results
        samp_res = pd.read_csv(
            get_data_path('context-subject-ordination.tsv'),
            sep='\t',
            index_col=0)
        feat_res = pd.read_csv(
            get_data_path('context-features-ordination.tsv'),
            sep='\t',
            index_col=0)
        bt_res = get_data_path('phylogenetic-table.biom')
        bt_res = load_table(bt_res)
        tree_res = get_data_path('expected-labeled-phylogeny.nwk')
        tree_res = TreeNode.read(tree_res,
                                 format='newick')
        # Read the expected results
        samp_exp = pd.read_csv(
            get_data_path('expected-phylo-context-subject-ordination.tsv'),
            sep='\t',
            index_col=0)
        feat_exp = pd.read_csv(
            get_data_path('expected-phylo-context-features-ordination.tsv'),
            sep='\t',
            index_col=0)
        bt_exp = load_table(get_data_path('expected-phylo-table.biom'))
        tree_exp = get_data_path('expected-labeled-phylogeny.nwk')
        tree_exp = TreeNode.read(tree_exp,
                                 format='newick')
        # Check that the distance matrix matches our expectations
        comp_col = ['PC1', 'PC2', 'PC3']
        cent_ = samp_res[comp_col].mean().values.max()
        self.assertAlmostEqual(cent_, 0)
        cent_ = feat_res[comp_col].mean().values.max()
        self.assertAlmostEqual(cent_, 0)
        # check renamed names are consistent
        name_check_ = [x.name == y.name for x, y in zip(tree_res.postorder(),
                                                        tree_exp.postorder())]
        name_check_ = all(name_check_)
        self.assertEqual(name_check_, True)
        # check matched arrays
        assert_allclose(absolute_sort(samp_res[comp_col].values),
                        absolute_sort(samp_exp[comp_col].values),
                        atol=.5)
        assert_allclose(absolute_sort(feat_res[comp_col].values),
                        absolute_sort(feat_exp[comp_col].values),
                        atol=.5)
        assert_array_almost_equal(bt_res.matrix_data.toarray(),
                                  bt_exp.matrix_data.toarray())


if __name__ == "__main__":
    unittest.main()
